# -*- coding: utf-8 -*-
import httpretty
from os import path
from test_auth import FakeAuth

from pybitbucket.bitbucket import Bitbucket
from pybitbucket.bitbucket import Client


class TestEntrypoints(object):
    @classmethod
    def setup_class(cls):
        cls.test_dir, current_file = path.split(path.abspath(__file__))
        cls.client = Client(FakeAuth())

    @httpretty.activate
    def test_find_current_user(self):
        url = ('https://api.bitbucket.org/2.0/user')
        example_path = path.join(self.test_dir, 'User.json')
        with open(example_path) as f:
            example = f.read()
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        user = next(Bitbucket(client=self.client).userForMyself())
        assert 'evzijst' == user.username
        assert 'Erik van Zijst' == user.display_name

    @httpretty.activate
    def test_find_user_by_username(self):
        url = ('https://api.bitbucket.org/2.0/users/evzijst')
        example_path = path.join(self.test_dir, 'User.json')
        with open(example_path) as f:
            example = f.read()
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        user = next(Bitbucket(client=self.client).userByUsername(
            username='evzijst'))
        assert 'evzijst' == user.username
        assert 'Erik van Zijst' == user.display_name



# -*- coding: utf-8 -*-
import httpretty
import json
from os import path
from test_auth import FakeAuth

from pybitbucket.team import Team
from pybitbucket.team import TeamRole
from pybitbucket.bitbucket import Client


class TestTeam(object):
    @classmethod
    def setup_class(cls):
        cls.test_dir, current_file = path.split(path.abspath(__file__))
        cls.client = Client(FakeAuth())

    def test_team_string_representation(self):
        example_path = path.join(self.test_dir, 'example_single_team.json')
        with open(example_path) as f:
            example = json.load(f)
        team = Team(example, client=self.client)
        # Just tests that the __str__ method works and
        # that it does not use the default representation
        team_str = "%s" % team
        print(team_str)
        assert not team_str.startswith('<')
        assert not team_str.endswith('>')
        assert team_str.startswith('Team username:')

    @httpretty.activate
    def test_team_list(self):
        url1 = ('https://api.bitbucket.org/2.0/teams?role=admin')
        path1 = path.join(
            self.test_dir,
            'example_teams_page_1.json')
        with open(path1) as example1_file:
            example1 = example1_file.read()
        httpretty.register_uri(
            httpretty.GET,
            url1,
            content_type='application/json',
            body=example1,
            status=200)

        teams = Team.find_teams_for_role(
            TeamRole.ADMIN,
            client=self.client)
        team_list = list()
        team_list.append(next(teams))
        team_list.append(next(teams))
        team_list.append(next(teams))

        url2 = (
            'https://' +
            'staging.bitbucket.org/api' +
            '/2.0/teams?role=admin&page=2')
        path2 = path.join(
            self.test_dir,
            'example_teams_page_2.json')
        with open(path2) as example2_file:
            example2 = example2_file.read()
        httpretty.register_uri(
            httpretty.GET,
            url2,
            content_type='application/json',
            body=example2,
            status=200)

        for t in teams:
            team_list.append(t)
        s = "%s" % team_list[0]
        assert s.startswith('Team username:')
        assert 5 == len(team_list)

    @httpretty.activate
    def test_find_team_by_username(self):
        url = ('https://api.bitbucket.org/2.0/teams/teamsinspace')
        example_path = path.join(self.test_dir, 'example_single_team.json')
        with open(example_path) as f:
            example = f.read()
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        team = Team.find_team_by_username('teamsinspace', client=self.client)
        assert 'teamsinspace' == team.username
        assert 'Teams In Space' == team.display_name

    @httpretty.activate
    def test_members_link(self):
        url = ('https://api.bitbucket.org/2.0/teams/teamsinspace')
        example_path = path.join(self.test_dir, 'example_single_team.json')
        with open(example_path) as f:
            example = f.read()
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        team = Team.find_team_by_username('teamsinspace', client=self.client)

        url = ('https://' +
               'api.bitbucket.org' +
               '/2.0/teams/teamsinspace/members')
        example_path = path.join(self.test_dir, 'example_members.json')
        with open(example_path) as f:
            example = f.read()
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        member = next(team.members())
        assert 'mbertrand80' == member.username
        assert 'Marcus Bertrand' == member.display_name



# -*- coding: utf-8 -*-

from __future__ import print_function

import os
import sys
import time
import subprocess

# Import parameters from the setup file.
sys.path.append('.')
from setup import (
    setup_dict, get_project_files, print_success_message,
    print_failure_message, _lint, _test, _test_all,
    CODE_DIRECTORY, DOCS_DIRECTORY, TESTS_DIRECTORY, PYTEST_FLAGS)

from paver.easy import options, task, needs, consume_args
from paver.setuputils import install_distutils_tasks
from contextlib import contextmanager
from io import BytesIO

options(setup=setup_dict)

install_distutils_tasks()

# Miscellaneous helper functions


def print_passed():
    # generated on http://patorjk.com/software/taag/#p=display&f=Small&t=PASSED
    print_success_message(r'''  ___  _   ___ ___ ___ ___
 | _ \/_\ / __/ __| __|   \
 |  _/ _ \\__ \__ \ _|| |) |
 |_|/_/ \_\___/___/___|___/
''')


def print_failed():
    # generated on http://patorjk.com/software/taag/#p=display&f=Small&t=FAILED
    print_failure_message(r'''  ___ _   ___ _    ___ ___
 | __/_\ |_ _| |  | __|   \
 | _/ _ \ | || |__| _|| |) |
 |_/_/ \_\___|____|___|___/
''')


class cwd(object):
    """Class used for temporarily changing directories. Can be though of
    as a `pushd /my/dir' then a `popd' at the end.
    """
    def __init__(self, newcwd):
        """:param newcwd: directory to make the cwd
        :type newcwd: :class:`str`
        """
        self.newcwd = newcwd

    def __enter__(self):
        self.oldcwd = os.getcwd()
        os.chdir(self.newcwd)
        return os.getcwd()

    def __exit__(self, type_, value, traceback):
        # This acts like a `finally' clause: it will always be executed.
        os.chdir(self.oldcwd)


# Task-related functions

def _doc_make(*make_args):
    """Run make in sphinx' docs directory.

    :return: exit code
    """
    if sys.platform == 'win32':
        # Windows
        make_cmd = ['make.bat']
    else:
        # Linux, Mac OS X, and others
        make_cmd = ['make']
    make_cmd.extend(make_args)

    # Account for a stupid Python "bug" on Windows:
    # <http://bugs.python.org/issue15533>
    with cwd(DOCS_DIRECTORY):
        retcode = subprocess.call(make_cmd)
    return retcode

@contextmanager
def stdout_redirector(stream):
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout

def git_check_master():
    current_branch = subprocess.check_output(
        ['git', 'rev-parse', '--abbrev-ref', 'HEAD']).strip()
    if current_branch != 'master':
        print_failure_message(
            'The current branch is {}. '
            'Release tasks can only be performed on master. '
            'Use `git checkout master` to switch. '
            .format(current_branch))
        raise SystemExit(1)

def git_check_remote():
    local = subprocess.check_output(
        ['git', 'rev-parse', '@'])
    remote = subprocess.check_output(
        ['git', 'rev-parse', '@{u}'])
    base = subprocess.check_output(
        ['git', 'merge-base', '@', '@{u}'])
    if local == remote:
        print('Local and remote are the same.')
    elif local == base:
        print_failure_message(
            'The current branch is behind the remote. '
            'Use `git pull` to bring the local branch current. ')
        raise SystemExit(1)
    elif remote == base:
        print_failure_message(
            'The current branch is ahead of the remote. '
            'Use `git push` to bring the remote branch current. ')
        raise SystemExit(1)
    else:
        print_failure_message(
            'The current branch has diverged from remote. '
            'Use `git pull` to bring the remote branch current, '
            'then merge and use `git push` to synchronize branches.')
        raise SystemExit(1)

def version_bump(part):
    try:
        import bumpversion
    except ImportError:
        print_failure_message(
            'Install bumpversion to use this task, '
            "i.e., `pip install --upgrade bumpversion`.")
        raise SystemExit(1)
    bumpversion.main([part, '--commit', '--tag', '--list'])

def release(part):
    # Perform any pre-flight checks
    git_check_master()
    git_check_remote()
    # _test_all() returns the number of failed tests,
    # so when _test_all() is true, there were failed tests.
    if _test_all():
        print_failure_message(
            'Cannot release if tests do not pass.')
        raise SystemExit(1)
    # Bump the version, create a bump commit, and tag
    version_bump(part)
    version = subprocess.check_output(
        ['git', 'describe', '--abbrev=0', '--tags']).strip()
    print(
        'Created new commit and tag for version bump. '
        'Use `git reset --hard HEAD~1` to rollback the commit, '
        'and `git tag -d {}` to rollback the tag.'
        .format(version))
    # Reload setup_dict to pickup new metadata.version
    reload(setup)
    # Build the pip package, upload to PyPI, and push
    print(
        'Begin packaging and uploading for {}'.
        format(metadata.version))
    upload()
    subprocess.check_call(['git', 'push'])
    subprocess.check_call(['git', 'push', '--tags'])


# Tasks

@task
def dep_install():
    """Install or upgrade development dependencies."""
    try:
        import pip
    except ImportError:
        print_failure_message(
            'Install pip to use this task, '
            "i.e., `sudo apt-get install python-pip`.")
        raise SystemExit(1)

    pip.main(['install', '--upgrade', '-r', 'requirements-dev.txt'])


@task
@needs('dep_install')
def prepare():
    """ Prepare complete environment """
    sh("python setup.py develop")


@task
def dep_check():
    """Check if any installed dependencies have newer versions."""
    try:
        import pip
    except ImportError:
        print_failure_message(
            'Install pip to use this task, '
            "i.e., `sudo apt-get install python-pip`.")
        raise SystemExit(1)
    f = BytesIO()
    with stdout_redirector(f):
        pip.main(['list', '--outdated'])
    if not f.getvalue():
        print_passed()
    else:
        print('Outdated pip dependencies:')
        print('{0}'.format(f.getvalue()))
        print_failed()
    raise SystemExit(len(f.getvalue()))


@task
@needs('doc_html', 'setuptools.command.sdist')
def sdist():
    """Build the HTML docs and the tarball."""
    pass

@task
@needs('sdist', 'distutils.command.upload')
def upload():
    """Upload package to PyPI."""
    pass

@task
def test():
    """Run the unit tests."""
    raise SystemExit(_test())


@task
def lint():
    # This refuses to format properly when running `paver help' unless
    # this ugliness is used.
    ('Perform PEP8 style check, run PyFlakes, and run McCabe complexity '
     'metrics on the code.')
    raise SystemExit(_lint())


@task
def test_all():
    """Perform a style check and run all unit tests."""
    retcode = _test_all()
    if retcode == 0:
        print_passed()
    else:
        print_failed()
    raise SystemExit(retcode)


@task
@consume_args
def run(args):
    """Run the package's main script. All arguments are passed to it."""
    # The main script expects to get the called executable's name as
    # argv[0]. However, paver doesn't provide that in args. Even if it did (or
    # we dove into sys.argv), it wouldn't be useful because it would be paver's
    # executable. So we just pass the package name in as the executable name,
    # since it's close enough. This should never be seen by an end user
    # installing through Setuptools anyway.
    from pybitbucket.main import main
    raise SystemExit(main([CODE_DIRECTORY] + args))


@task
def commit():
    """Commit only if all the tests pass."""
    if _test_all() == 0:
        subprocess.check_call(['git', 'commit'])
    else:
        print_failure_message('\nTests failed, not committing.')


@task
def release_major():
    """Test, package, and release a major version change."""
    release('major')


@task
def release_minor():
    """Test, package, and release a minor version change."""
    release('minor')


@task
def release_patch():
    """Test, package, and release a patch version change."""
    release('patch')


@task
def coverage():
    """Run tests and show test coverage report."""
    try:
        import pytest_cov  # NOQA
    except ImportError:
        print_failure_message(
            'Install the pytest coverage plugin to use this task, '
            "i.e., `pip install pytest-cov'.")
        raise SystemExit(1)
    import pytest
    pytest.main(PYTEST_FLAGS + [
        '--cov', CODE_DIRECTORY,
        '--cov-report', 'term-missing',
        '--cov-report', 'xml',
        TESTS_DIRECTORY])


@task  # NOQA
def doc_watch():
    """Watch for changes in the docs and rebuild HTML docs when changed."""
    try:
        from watchdog.events import FileSystemEventHandler
        from watchdog.observers import Observer
    except ImportError:
        print_failure_message(
            'Install the watchdog package to use this task, '
            "i.e., `pip install watchdog`.")
        raise SystemExit(1)

    class RebuildDocsEventHandler(FileSystemEventHandler):
        def __init__(self, base_paths):
            self.base_paths = base_paths

        def dispatch(self, event):
            """Dispatches events to the appropriate methods.
            :param event: The event object representing the file system event.
            :type event: :class:`watchdog.events.FileSystemEvent`
            """
            for base_path in self.base_paths:
                if event.src_path.endswith(base_path):
                    super(RebuildDocsEventHandler, self).dispatch(event)
                    # We found one that matches. We're done.
                    return

        def on_modified(self, event):
            print_failure_message('Modification detected. Rebuilding docs.')
            # # Strip off the path prefix.
            # import os
            # if event.src_path[len(os.getcwd()) + 1:].startswith(
            #         CODE_DIRECTORY):
            #     # sphinx-build doesn't always pick up changes on code files,
            #     # even though they are used to generate the documentation. As
            #     # a workaround, just clean before building.
            doc_html()
            print_success_message('Docs have been rebuilt.')

    print_success_message(
        'Watching for changes in project files, press Ctrl-C to cancel...')
    handler = RebuildDocsEventHandler(get_project_files())
    observer = Observer()
    observer.schedule(handler, path='.', recursive=True)
    observer.start()
    try:
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        observer.stop()
        observer.join()


@task
@needs('doc_html')
def doc_open():
    """Build the HTML docs and open them in a web browser."""
    doc_index = os.path.join(DOCS_DIRECTORY, 'build', 'html', 'index.html')
    if sys.platform == 'darwin':
        # Mac OS X
        subprocess.check_call(['open', doc_index])
    elif sys.platform == 'win32':
        # Windows
        subprocess.check_call(['start', doc_index], shell=True)
    elif sys.platform == 'linux2':
        # All freedesktop-compatible desktops
        subprocess.check_call(['xdg-open', doc_index])
    else:
        print_failure_message(
            "Unsupported platform. Please open `{0}' manually.".format(
                doc_index))


@task
def get_tasks():
    """Get all paver-defined tasks."""
    from paver.tasks import environment
    for t in environment.get_tasks():
        print(t.shortname)


@task
def doc_html():
    """Build the HTML docs."""
    retcode = _doc_make('html')

    if retcode:
        raise SystemExit(retcode)


@task
def doc_clean():
    """Clean (delete) the built docs."""
    retcode = _doc_make('clean')

    if retcode:
        raise SystemExit(retcode)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""Project metadata

Information describing the project.
"""

# The package name, which is also the "UNIX name" for the project.
package = 'pybitbucket_fork'
project = "Fork of PyBitbucket"
project_no_spaces = project.replace(' ', '')
version = '0.12.1'
description = 'Fork of a Python wrapper for the Bitbucket API'
authors = ['Ian Buchanan', 'Bernard Pratz']
authors_string = ', '.join(authors)
emails = ['ibuchanan@atlassian.com', 'z+pybitbucket+pub@m0g.net']
license = 'Apache'
copyright = '2015 ' + 'Atlassian'
url = 'https://github.com/guyzmo/pybitbucket'
bugtrack_url = 'https://github.com/guyzmo/pybitbucket/issues'



# -*- coding: utf-8 -*-
import httpretty
from os import path
from test_auth import FakeAuth

from util import data_from_file
from pybitbucket.bitbucket import Client


class TestRemoteRelationships(object):

    def setup_class(cls):
        cls.test_dir, current_file = path.split(path.abspath(__file__))
        cls.client = Client(FakeAuth())

    @httpretty.activate
    def test_single_item(self):
        url = (
            'https://' +
            'api.bitbucket.org' +
            '/2.0/repositories/' +
            'teamsinspace/teamsinspace.bitbucket.org')
        example = data_from_file(
            self.test_dir,
            'Repository.json')
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        repo_list = list(self.client.remote_relationship(url))
        s = "%s" % repo_list[0]
        assert s.startswith('Repository full_name:')
        assert 1 == len(repo_list)

    @httpretty.activate
    def test_one_page_of_items(self):
        url = (
            'https://' +
            'api.bitbucket.org' +
            '/2.0/repositories')
        example = data_from_file(
            self.test_dir,
            'Repository_list.json')
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        repo_list = list(self.client.remote_relationship(url))
        s = "%s" % repo_list[0]
        assert s.startswith('Repository full_name:')
        assert 2 == len(repo_list)

    @httpretty.activate
    def test_two_pages_of_items(self):
        url1 = (
            self.client.get_bitbucket_url() +
            '/2.0/snippets' +
            '?role=owner')
        example1 = data_from_file(
            self.test_dir,
            'example_snippets_page_1.json')
        httpretty.register_uri(
            httpretty.GET,
            url1,
            match_querystring=True,
            content_type='application/json',
            body=example1,
            status=200)
        snip_iter = self.client.remote_relationship(url1)

        snippet_list = list()
        snippet_list.append(next(snip_iter))
        snippet_list.append(next(snip_iter))
        snippet_list.append(next(snip_iter))

        url2 = (
            'https://' +
            'staging.bitbucket.org/api' +
            '/2.0/snippets' +
            '?role=owner&page=2')
        example2 = data_from_file(
            self.test_dir,
            'example_snippets_page_2.json')
        httpretty.register_uri(
            httpretty.GET,
            url2,
            match_querystring=True,
            content_type='application/json',
            body=example2,
            status=200)

        for snip in snip_iter:
            snippet_list.append(snip)

        s = "%s" % snippet_list[0]
        assert s.startswith('Snippet id:')
        assert 5 == len(snippet_list)



# -*- coding: utf-8 -*-
from test_bitbucketbase import BitbucketFixture

import json
from uritemplate import expand
from pybitbucket.user import User, UserV1
from pybitbucket.bitbucket import Bitbucket

import httpretty


class UserFixture(BitbucketFixture):
    # GIVEN: a class under test
    class_under_test = 'User'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return User(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes a user
    username = 'evzijst'


class UserV1Fixture(BitbucketFixture):
    # GIVEN: a class under test
    class_under_test = 'UserV1'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return UserV1(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes a user
    username = 'ianbuchanan'


class TestGettingTheStringRepresentation(UserFixture):
    @classmethod
    def setup_class(cls):
        cls.user_str = str(cls.example_object())

    def test_string_is_not_the_default_format(self):
        assert not self.user_str.startswith('<')
        assert not self.user_str.endswith('>')

    def test_string_has_the_class_name_and_id_attribute(self):
        assert self.user_str.startswith('User username:')


class TestCheckingTheExampleData(UserFixture):
    @classmethod
    def setup_class(cls):
        cls.data = json.loads(cls.resource_data())

    def test_passes_the_type_check(self):
        assert User.is_type(self.data)


class TestCheckingTheExampleDataForV1(UserV1Fixture):
    @classmethod
    def setup_class(cls):
        cls.data = json.loads(cls.resource_data())

    def test_passes_the_type_check(self):
        assert UserV1.is_type(self.data)


class TestFindingUserByUsername(UserFixture):
    @httpretty.activate
    def test_response_is_a_user(self):
        httpretty.register_uri(
            httpretty.GET,
            self.resource_url(),
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = User.find_user_by_username(
            username=self.username,
            client=self.test_client)
        assert isinstance(response, User)
        assert 'evzijst' == response.username
        assert 'Erik van Zijst' == response.display_name


class TestFindingCurrentUser(UserFixture):
    @httpretty.activate
    def test_response_is_a_user(self):
        url = (
            Bitbucket(client=self.test_client)
            .data
            .get('_links', {})
            .get('userForMyself', {})
            .get('href'))
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = User.find_current_user(client=self.test_client)
        assert isinstance(response, User)
        assert 'evzijst' == response.username
        assert 'Erik van Zijst' == response.display_name


class TestCallingFollowersLink(UserFixture):
    @httpretty.activate
    def test_response_is_a_list_of_users(self):
        user = self.example_object()
        url = user.links.get('followers', {}).get('href')
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        assert list(user.followers())
        assert isinstance(next(user.followers()), User)


class TestCommonAttributesForV1andV2(UserV1Fixture):
    @httpretty.activate
    def test_attributes_are_not_empty(self):
        user = self.example_object()
        assert user.username
        assert user.display_name


class TestNavigatingToOAuthConsumers(UserFixture):
    @httpretty.activate
    def test_attributes_are_not_empty(self):
        from pybitbucket.consumer import Consumer
        user = self.example_object()
        url = user.v1.links.get('consumers', {}).get('href')
        example = self.data_from_file('Consumer_list.json')
        httpretty.register_uri(
            httpretty.GET,
            url,
            content_type='application/json',
            body=example,
            status=200)
        assert list(user.v1.consumers())
        assert isinstance(next(user.v1.consumers()), Consumer)


class TestNavigatingFromV1toV2(UserV1Fixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()
        user_template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('userByUsername', {})
            .get('href'))
        cls.user_url = expand(
            user_template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'username': cls.username,
            })
        cls.user_data = cls.resource_list_data('User')

    @httpretty.activate
    def test_v2_self_returns_a_user(self):
        httpretty.register_uri(
            httpretty.GET,
            self.user_url,
            content_type='application/json',
            body=self.user_data,
            status=200)
        response = self.response.v2.self()
        assert isinstance(response, User)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Defines the BuildStatus resource and registers the type with the Client.

Classes:
- BuildStatusStates: enumerates the possible states of a build status
- BuildStatusPayload: encapsulates payload for creating
    and modifying build status
- BuildStatus: represents the result of a build
"""

from uritemplate import expand
from voluptuous import Schema, Required, Optional, In, Url

from pybitbucket.bitbucket import (
    Bitbucket, BitbucketBase, Client, PayloadBuilder, Enum)


class BuildStatusStates(Enum):
    INPROGRESS = 'INPROGRESS'
    SUCCESSFUL = 'SUCCESSFUL'
    FAILED = 'FAILED'


class BuildStatusPayload(PayloadBuilder):
    """
    A builder object to help create payloads
    for creating and updating build statuses.
    """

    schema = Schema({
        Required('key'): str,
        Required('state'): In(list(BuildStatusStates)),
        Required('url'): Url(),
        Optional('name'): str,
        Optional('description'): str
    })

    def __init__(
            self,
            payload=None,
            owner=None,
            repository_name=None,
            revision=None):
        super(self.__class__, self).__init__(payload=payload)
        self._owner = owner
        self._repository_name = repository_name
        self._revision = revision

    @property
    def owner(self):
        return self._owner

    @property
    def repository_name(self):
        return self._repository_name

    @property
    def revision(self):
        return self._revision

    def add_owner(self, owner):
        return BuildStatusPayload(
            payload=self._payload.copy(),
            owner=owner,
            repository_name=self.repository_name,
            revision=self.revision)

    def add_repository_name(self, repository_name):
        return BuildStatusPayload(
            payload=self._payload.copy(),
            owner=self.owner,
            repository_name=repository_name,
            revision=self.revision)

    def add_revision(self, revision):
        return BuildStatusPayload(
            payload=self._payload.copy(),
            owner=self.owner,
            repository_name=self.repository_name,
            revision=revision)

    def add_name(self, name):
        new = self._payload.copy()
        new['name'] = name
        return BuildStatusPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name,
            revision=self.revision)

    def add_description(self, description):
        new = self._payload.copy()
        new['description'] = description
        return BuildStatusPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name,
            revision=self.revision)

    def add_key(self, key):
        new = self._payload.copy()
        new['key'] = key
        return BuildStatusPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name,
            revision=self.revision)

    def add_state(self, state):
        new = self._payload.copy()
        new['state'] = state
        return BuildStatusPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name,
            revision=self.revision)

    def add_url(self, url):
        new = self._payload.copy()
        new['url'] = url
        return BuildStatusPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name,
            revision=self.revision)


class BuildStatus(BitbucketBase):
    """Represents a build status."""

    id_attribute = 'key'
    resource_type = 'build'
    templates = {
        'create': (
            '{+bitbucket_url}' +
            '/2.0/repositories{/owner,repository_name}' +
            '/commit{/revision}/statuses/build')
    }

    @staticmethod
    def is_type(data):
        return (BuildStatus.has_v2_self_url(data))

    @classmethod
    def create(
            cls,
            payload,
            revision=None,
            repository_name=None,
            owner=None,
            client=None):
        """Create a new build status.

        :param payload: the options for creating the new build status.
        :type payload: BuildStatusPayload
        :param revision: revision in the repository,
            also known as commit. Optional, if provided in the payload.
        :type revision: str
        :param repository_name: name of the repository,
            also known as repo_slug. Optional, if provided in the payload.
        :type repository_name: str
        :param owner: the owner of the repository.
            If not provided as parameter, it may be provided in the payload.
            If neither, assume the current user.
        :type owner: str
        :param client: the configured connection to Bitbucket.
            If not provided, assumes an Anonymous connection.
        :type client: bitbucket.Client
        :returns: the new build status object.
        :rtype: BuildStatus
        :raises: ValueError
        """
        client = client or Client()
        owner = owner or payload.owner or client.get_username()
        repository_name = repository_name or payload.repository_name
        revision = revision or payload.revision
        if not (owner and repository_name and revision):
            raise ValueError(
                'owner, repository_name, and revision'
                ' are required')
        json = payload.validate().build()
        api_url = expand(
            cls.templates['create'], {
                'bitbucket_url': client.get_bitbucket_url(),
                'owner': owner,
                'repository_name': repository_name,
                'revision': revision
            })
        return cls.post(api_url, json=json, client=client)

    def modify(self, payload):
        """
        A convenience method for changing the current build status.
        """
        return self.put(json=payload.validate().build())

    @staticmethod
    def find_buildstatus_for_repository_commit_by_key(
            repository_name,
            revision,
            key,
            owner=None,
            client=None):
        """
        A convenience method for finding a specific build status.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a BuildStatus object, instead of the
        generator.
        """
        client = client or Client()
        owner = owner or client.get_username()
        return next(
            Bitbucket(client=client).repositoryCommitBuildStatusByKey(
                owner=owner,
                repository_name=repository_name,
                revision=revision,
                key=key))

    @staticmethod
    def find_buildstatuses_for_repository_commit(
            repository_name,
            revision,
            owner=None,
            client=None):
        """
        A convenience method for finding build statuses
        for a repository's commit.
        The method is a generator BuildStatus objects.
        """
        client = client or Client()
        owner = owner or client.get_username()
        return Bitbucket(client=client).repositoryCommitBuildStatuses(
            owner=owner,
            repository_name=repository_name,
            revision=revision)


Client.bitbucket_types.add(BuildStatus)



# -*- coding: utf-8 -*-
import json
from os import path
from test_auth import FakeAuth

from util import data_from_file
from pybitbucket.bitbucket import Client
from pybitbucket.commit import Commit
from pybitbucket.hook import Hook
from pybitbucket.repository import Repository
from pybitbucket.snippet import Snippet
from pybitbucket.team import Team
from pybitbucket.user import User


class TestEmbeddedResources(object):

    def setup_class(cls):
        cls.test_dir, current_file = path.split(path.abspath(__file__))
        cls.client = Client(FakeAuth())

        example_commit = json.loads(
                data_from_file(
                    cls.test_dir,
                    'Commit.json'))
        cls.commit = cls.client.convert_to_object(example_commit)

    def test_webhook_subject_is_a_repository(self):
        example = json.loads(
                data_from_file(
                    self.test_dir,
                    'Hook.json'))
        my_hook = self.client.convert_to_object(example)
        assert isinstance(my_hook, Hook)
        assert isinstance(my_hook.subject, Repository)

    def test_commit_author_is_a_user(self):
        assert isinstance(self.commit, Commit)
        assert isinstance(self.commit.author, User)

    def test_commit_repository_is_a_repository(self):
        assert isinstance(self.commit, Commit)
        assert isinstance(self.commit.parents[0], Commit)

    def test_commit_parent_is_another_commit(self):
        assert isinstance(self.commit, Commit)
        assert isinstance(self.commit.parents[0], Commit)

    def test_repository_owner_is_a_team(self):
        example = json.loads(
                data_from_file(
                    self.test_dir,
                    'Repository.json'))
        my_repo = self.client.convert_to_object(example)
        assert isinstance(my_repo, Repository)
        assert isinstance(my_repo.owner, Team)

    def test_snippet_creator_and_owner_are_users(self):
        example = json.loads(
                data_from_file(
                    self.test_dir,
                    'Snippet.json'))
        my_snip = self.client.convert_to_object(example)
        assert isinstance(my_snip, Snippet)
        assert isinstance(my_snip.owner, User)
        assert isinstance(my_snip.creator, User)



# -*- coding: utf-8 -*-
from test_bitbucketbase import BitbucketFixture
import json

import httpretty
from uritemplate import expand
from pybitbucket.bitbucket import Bitbucket
from pybitbucket.branchrestriction import (
    BranchRestriction, BranchRestrictionKind, BranchRestrictionPayload)
from pybitbucket.user import User
from voluptuous import MultipleInvalid


class BranchRestrictionFixture(BitbucketFixture):
    # GIVEN: a class under test
    class_under_test = 'BranchRestriction'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return BranchRestriction(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: The URL for posting branch-restriction resources
    @classmethod
    def resource_list_url(cls):
        bitbucket = Bitbucket(cls.test_client)
        t = bitbucket.data['_links']['repositoryBranchRestrictions']['href']
        url = expand(
            t, {
                'owner': cls.owner,
                'repository_name': cls.repository_name,
            })
        return url

    # GIVEN: Example data attributes for a build status
    owner = 'ianbuchanan'
    repository_name = 'example'
    restriction_id = 913351
    kind = BranchRestrictionKind.PUSH
    user = 'ian_buchanan'
    users = [user, 'tpettersen']
    group_username = user
    group_name = 'developer-relations'
    pattern = 'feature/*'


class BranchRestrictionPayloadFixture(BranchRestrictionFixture):
    builder = BranchRestrictionPayload()

    # GIVEN: a class under test
    class_under_test = 'BranchRestrictionPayload'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return BranchRestrictionPayload(json.loads(cls.resource_data()))


class TestGettingTheStringRepresentation(BranchRestrictionFixture):
    @classmethod
    def setup_class(cls):
        cls.branchrestriction_str = str(cls.example_object())

    def test_string_is_not_the_default_format(self):
        assert not self.branchrestriction_str.startswith('<')
        assert not self.branchrestriction_str.endswith('>')

    def test_string_has_the_class_name_and_id_attribute(self):
        assert self.branchrestriction_str.startswith('BranchRestriction id:')


class TestCreatingPushPayloadWithPatternAndUsers(
        BranchRestrictionPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = BranchRestrictionPayload() \
            .add_kind(BranchRestrictionKind.PUSH) \
            .add_pattern('master') \
            .add_user_by_username('ibuchanan')

    def test_payload_structure(self):
        assert self.payload.validate().build() == {
            "kind": "push",
            "pattern": "master",
            "users": [{
                    "username": "ibuchanan"
                }]
            }


class TestCreatingNewBranchRestriction(BranchRestrictionFixture):
    @classmethod
    def setup_class(cls):
        cls.url = expand(
            BranchRestriction.templates['create'], {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.owner,
                'repository_name': cls.repository_name
            })

    @httpretty.activate
    def test_response_is_a_branchrestriction(self):
        print(self.url)
        httpretty.register_uri(
            httpretty.POST,
            self.url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        payload = BranchRestrictionPayload() \
            .add_kind(self.kind) \
            .add_pattern(self.pattern) \
            .add_user_by_username(self.user) \
            .add_owner(self.owner) \
            .add_repository_name(self.repository_name)
        response = BranchRestriction.create(payload, client=self.test_client)
        assert 'application/json' == \
            httpretty.last_request().headers.get('Content-Type')
        assert isinstance(response, BranchRestriction)


class TestUpdatingBranchRestriction(BranchRestrictionFixture):
    @httpretty.activate
    def test_response_is_a_branchrestriction(self):
        httpretty.register_uri(
            httpretty.PUT,
            self.resource_url(),
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        payload = BranchRestrictionPayload() \
            .add_kind(BranchRestrictionKind.PUSH) \
            .add_pattern('developing')
        response = self.example_object().modify(payload)
        assert 'application/json' == \
            httpretty.last_request().headers.get('Content-Type')
        assert isinstance(response, BranchRestriction)


class TestFindingBranchRestrictions(BranchRestrictionFixture):
    @httpretty.activate
    def test_response_is_a_branchrestriction_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.resource_list_url(),
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = BranchRestriction.find_branchrestrictions_for_repository(
            owner=self.owner,
            repository_name=self.repository_name)
        assert isinstance(next(response), BranchRestriction)


class TestFindingBranchRestrictionById(BranchRestrictionFixture):
    @httpretty.activate
    def test_response_is_a_branchrestriction(self):
        httpretty.register_uri(
            httpretty.GET,
            self.resource_url(),
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = \
            BranchRestriction.find_branchrestriction_for_repository_by_id(
                owner=self.owner,
                repository_name=self.repository_name,
                restriction_id=self.restriction_id)
        assert isinstance(response, BranchRestriction)


class TestDeletingBranchRestriction(BranchRestrictionFixture):
    @httpretty.activate
    def test_response_is_not_an_exception(self):
        httpretty.register_uri(
            httpretty.DELETE,
            self.resource_url(),
            status=204)
        result = self.example_object().delete()
        assert result is None


class TestCreatingDefaultBranchRestrictionPayload(
        BranchRestrictionPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.data = cls.builder.build()

    def test_default_payload_is_empty_dictionary(self):
        assert {} == self.data

    def test_default_is_invalid(self):
        try:
            self.builder.validate()
            assert False
        except Exception as e:
            assert isinstance(e, MultipleInvalid)


class TestAddingOwnerToBranchRestrictionPayload(
        BranchRestrictionPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_owner = cls.builder.add_owner(cls.owner)

    def test_immutability_after_add(self):
        assert self.with_owner
        assert {} == self.builder.build()

    def test_structure(self):
        assert self.owner == self.with_owner.owner


class TestAddingUsersToBranchRestrictionPayload(
        BranchRestrictionPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_user = \
            cls.builder.add_user_by_username(
                cls.user)
        cls.with_users = \
            cls.builder.add_users_from_usernames(
                cls.users)
        cls.user_obj = User(
            json.loads(cls.resource_data('User')),
            client=cls.test_client)
        cls.with_user_obj = cls.builder.add_user(cls.user_obj)

    def test_immutability_on_adding_users(self):
        assert self.with_user
        assert self.with_users
        assert {} == self.builder.build()

    def test_users_structure_for_one(self):
        expected = {
            'users': [
                {'username': self.user}
            ]
        }
        assert expected == self.with_user.build()

    def test_users_structure_for_many(self):
        expected = {'users': [{'username': u} for u in self.users]}
        assert expected == self.with_users.build()

    def test_users_structure_for_user_obj(self):
        expected = {
            'users': [
                {'username': 'evzijst'}
            ]
        }
        assert expected == self.with_user_obj.build()


class TestCreatingMinimalBranchRestrictionPayload(
        BranchRestrictionPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = BranchRestrictionPayload() \
            .add_kind(cls.kind) \
            .add_pattern(cls.pattern)
        cls.expected = json.loads(cls.resource_data(
            'BranchRestrictionPayload.minimal'))

    def test_minimum_viable_payload_structure_for_create(self):
        assert self.payload.validate().build() == self.expected


class TestCreatingFullBranchRestrictionPayload(
        BranchRestrictionPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = BranchRestrictionPayload() \
            .add_kind(BranchRestrictionKind.DELETE) \
            .add_pattern(cls.pattern) \
            .add_user_by_username(cls.user) \
            .add_group_by_username_and_groupname(
                cls.group_username,
                cls.group_name)
        cls.expected = json.loads(cls.resource_data(
            'BranchRestrictionPayload.full'))

    def test_full_payload_structure(self):
        assert self.payload.validate().build() == self.expected



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""A Python wrapper for the Bitbucket API"""

from pybitbucket import metadata

__version__ = metadata.version
__author__ = metadata.authors[0]
__license__ = metadata.license
__copyright__ = metadata.copyright



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Defines the BranchRestriction resource and registers the type with the Client.

Classes:
- BranchRestrictionKind: enumerates the possible restrictions for a branch.
- BranchRestrictionPayload: encapsulates payload for creating
    and modifying branch restrictions.
- BranchRestriction: represents a restriction on a branch for a repository.
"""

from uritemplate import expand
from voluptuous import Schema, Required, Optional, In, Invalid

from pybitbucket.bitbucket import (
    Bitbucket, BitbucketBase, Client, PayloadBuilder, Enum)


class BranchRestrictionKind(Enum):
    PUSH = 'push'
    DELETE = 'delete'
    FORCE = 'force'


class BranchRestrictionPayload(PayloadBuilder):
    """
    A builder object to help create payloads
    for creating and updating branch restrictions.
    """

    schema = Schema({
        Required('kind'): In(BranchRestrictionKind),
        Optional('pattern'): str,
        Optional('groups'): [{
            Required('owner'):
                {Required('username'): str},
            Required('slug'): str,
        }],
        Optional('users'): [{Required('username'): str}],
    })

    def __init__(
            self,
            payload=None,
            owner=None,
            repository_name=None):
        super(BranchRestrictionPayload, self).__init__(payload=payload)
        self._owner = owner
        self._repository_name = repository_name

    @property
    def owner(self):
        return self._owner

    @property
    def repository_name(self):
        return self._repository_name

    def add_owner(self, owner):
        return BranchRestrictionPayload(
            payload=self._payload.copy(),
            owner=owner,
            repository_name=self.repository_name)

    def add_repository_name(self, repository_name):
        return BranchRestrictionPayload(
            payload=self._payload.copy(),
            owner=self.owner,
            repository_name=repository_name)

    def add_kind(self, kind):
        new = self._payload.copy()
        new['kind'] = kind
        return BranchRestrictionPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def add_pattern(self, pattern):
        new = self._payload.copy()
        new['pattern'] = pattern
        return BranchRestrictionPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    # TODO: implement Group resource
    def add_group(self, group):
        return self.add_group_by_username_and_groupname(
            group.owner.username,
            group.name)

    def add_group_by_username_and_groupname(self, username, groupname):
        new = self._payload.copy()
        groups = self._payload.get('groups', [])
        groups.append({
            'owner': {'username': username},
            'slug': groupname})
        new['groups'] = groups
        return BranchRestrictionPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def add_user(self, user):
        return self.add_user_by_username(user.username)

    def add_user_by_username(self, username):
        new = self._payload.copy()
        users = self._payload.get('users', [])
        users.append({'username': username})
        new['users'] = users
        return BranchRestrictionPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def add_users_from_usernames(self, usernames):
        new = self._payload.copy()
        users = self._payload.get('users', [])
        for username in usernames:
            if {'username': username} not in users:
                users.append({'username': username})
        new['users'] = users
        return BranchRestrictionPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)


class BranchRestriction(BitbucketBase):
    id_attribute = 'id'
    resource_type = 'branch-restrictions'
    templates = {
        'create': (
            '{+bitbucket_url}' +
            '/2.0/repositories' +
            '{/owner,repository_name}' +
            '/branch-restrictions')
    }

    @staticmethod
    def is_type(data):
        return (BranchRestriction.has_v2_self_url(data))

    @classmethod
    def create(
            cls,
            payload,
            repository_name=None,
            owner=None,
            client=None):
        """Create a new branch-restriction.

        :param payload: the options for creating the new Branch Restriction
        :type payload: BranchRestrictionPayload
        :param repository_name: name of the destination repository,
            also known as repo_slug. Optional, if provided in the payload.
        :type repository_name: str
        :param owner: the owner of the destination repository.
            If not provided, assumes the current user.
        :type owner: str
        :param client: the configured connection to Bitbucket.
            If not provided, assumes an Anonymous connection.
        :type client: bitbucket.Client
        :returns: the new BranchRestriction object.
        :rtype: BranchRestriction
        :raises: MultipleInvalid, Invalid
        """
        client = client or Client()
        owner = owner or payload.owner
        repository_name = repository_name or payload.repository_name
        if not (owner and repository_name):
            raise Invalid('owner and repository_name are required')
        json = payload.validate().build()
        api_url = expand(
            cls.templates['create'], {
                'bitbucket_url': client.get_bitbucket_url(),
                'owner': owner,
                'repository_name': repository_name,
            })
        return cls.post(api_url, json=json, client=client)

    def modify(self, payload):
        """
        A convenience method for changing the current branch-restriction.
        The parameters make it easier to know what can be changed.
        """
        json = payload.validate().build()
        return self.put(json=json)

    @staticmethod
    def find_branchrestrictions_for_repository(
            repository_name,
            owner=None,
            client=None):
        """
        A convenience method for finding branch-restrictions for a repository.
        The method is a generator BranchRestriction objects.
        """
        client = client or Client()
        owner = owner or client.get_username()
        return Bitbucket(client=client).repositoryBranchRestrictions(
            owner=owner,
            repository_name=repository_name)

    @staticmethod
    def find_branchrestriction_for_repository_by_id(
            repository_name,
            restriction_id,
            owner=None,
            client=None):
        """
        A convenience method for finding a specific branch-restriction.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a BranchRestriction object, instead of the
        generator.
        """
        client = client or Client()
        owner = owner or client.get_username()
        return next(
            Bitbucket(
                client=client).repositoryBranchRestrictionByRestrictionId(
                    owner=owner,
                    repository_name=repository_name,
                    restriction_id=restriction_id))


Client.bitbucket_types.add(BranchRestriction)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Defines the Hook resource and registers the type with the Client.

Classes:
- HookEvent: enumerates the possible events for subscription.
- HookPayload: encapsulates payload for creating
    and modifying hooks.
- Hook: represents a web hook for a repository
"""

from uritemplate import expand
from voluptuous import Schema, Required, Optional, In

from pybitbucket.bitbucket import (
    Bitbucket, BitbucketBase, Client, PayloadBuilder, Enum)


class HookEvent(Enum):
    REPOSITORY_PUSH = 'repo:push'
    REPOSITORY_FORK = 'repo:fork'
    REPOSITORY_UPDATED = 'repo:updated'
    REPOSITORY_COMMIT_COMMENT_CREATED = 'repo:commit_comment_created'
    REPOSITORY_BUILD_STATUS_CREATED = 'repo:commit_status_created'
    REPOSITORY_BUILD_STATUS_UPDATE = 'repo:commit_status_updated'
    ISSUE_CREATED = 'issue:created'
    ISSUE_UPDATED = 'issue:updated'
    ISSUE_COMMENT_CREATED = 'issue:comment_created'
    PULL_REQUEST_CREATED = 'pullrequest:created'
    PULL_REQUEST_UPDATED = 'pullrequest:updated'
    PULL_REQUEST_APPROVED = 'pullrequest:approved'
    PULL_REQUEST_APPROVAL_REMOVED = 'pullrequest:unapproved'
    PULL_REQUEST_MERGED = 'pullrequest:fulfilled'
    PULL_REQUEST_DECLINED = 'pullrequest:rejected'
    PULL_REQUEST_COMMENT_CREATED = 'pullrequest:comment_created'
    PULL_REQUEST_COMMENT_UPDATED = 'pullrequest:comment_updated'
    PULL_REQUEST_COMMENT_DELETED = 'pullrequest:comment_deleted'


class HookPayload(PayloadBuilder):
    """
    A builder object to help create payloads
    for creating and updating hooks.
    """

    schema = Schema({
        Required('description'): str,
        Required('url'): str,
        Optional('active'): bool,
        Optional('events'): [In(HookEvent)],
        # Undocumented attributes
        Optional('skip_cert_verification'): bool,
    })

    def __init__(
            self,
            payload=None,
            owner=None,
            repository_name=None):
        super(self.__class__, self).__init__(payload=payload)
        self._owner = owner
        self._repository_name = repository_name

    @property
    def owner(self):
        return self._owner

    @property
    def repository_name(self):
        return self._repository_name

    def add_owner(self, owner):
        return HookPayload(
            payload=self._payload.copy(),
            owner=owner,
            repository_name=self._repository_name)

    def add_repository_name(self, name):
        return HookPayload(
            payload=self._payload.copy(),
            owner=self._owner,
            repository_name=name)

    def add_repository_full_name(self, full_name):
        owner, name = full_name.split('/', 1)
        return HookPayload(
            payload=self._payload.copy(),
            owner=owner,
            repository_name=name)

    def add_description(self, description):
        new = self._payload.copy()
        new['description'] = description
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def add_callback_url(self, callback_url):
        new = self._payload.copy()
        new['url'] = callback_url
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def activate(self):
        new = self._payload.copy()
        new['active'] = True
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def deactivate(self):
        new = self._payload.copy()
        new['active'] = False
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def add_event(self, event):
        new = self._payload.copy()
        new_events = self._payload.get('events', [])
        if event not in new_events:
            new_events.append(event)
        new['events'] = new_events
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def add_events(self, events):
        new = self._payload.copy()
        new_events = self._payload.get('events', [])
        for event in events:
            if event not in new_events:
                new_events.append(event)
        new['events'] = new_events
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def enable_cert_verification(self):
        new = self._payload.copy()
        new['skip_cert_verification'] = False
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)

    def disable_cert_verification(self):
        new = self._payload.copy()
        new['skip_cert_verification'] = True
        return HookPayload(
            payload=new,
            owner=self.owner,
            repository_name=self.repository_name)


class Hook(BitbucketBase):
    id_attribute = 'uuid'
    resource_type = 'hooks'
    templates = {
        'create': (
            '{+bitbucket_url}' +
            '/2.0/repositories' +
            '{/owner,repository_name}' +
            '/hooks')
    }

    @staticmethod
    def is_type(data):
        return (Hook.has_v2_self_url(data))

    @classmethod
    def create(
            cls,
            payload,
            repository_name=None,
            owner=None,
            client=None):
        """Create a new hook.

        :param payload: the options for creating the new hook
        :type payload: HookPayload
        :param repository_name: name of the repository,
            also known as repo_slug. Optional, if provided in the payload.
        :type repository_name: str
        :param owner: the owner of the repository.
            Optional, if provided in the payload.
            If not explicit as parameter or available in payload,
            uses the current user.
        :type owner: str
        :param client: the configured connection to Bitbucket.
            If not provided, assumes an Anonymous connection.
        :type client: bitbucket.Client
        :returns: the new hook object.
        :rtype: Hook
        :raises: MultipleInvalid
        """
        client = client or Client()
        owner = (
            owner or
            payload.owner or
            client.get_username())
        repository_name = (
            repository_name or
            payload.repository_name)
        if not (owner and repository_name):
            raise ValueError('owner and repository_name are required')
        json = payload.validate().build()
        api_url = expand(
            cls.templates['create'], {
                'bitbucket_url': client.get_bitbucket_url(),
                'owner': owner,
                'repository_name': repository_name,
            })
        return cls.post(api_url, json=json, client=client)

    def update(
            self,
            payload):
        """Update current hook.

        :param payload: the options for updating the new hook
        :type payload: HookPayload
        :returns: the new hook object.
        :rtype: Hook
        :raises: MultipleInvalid
        """
        return self.put(json=payload.validate().build())

    @staticmethod
    def find_hook_by_uuid_in_repository(
            uuid,
            repository_name,
            owner=None,
            client=None):
        """
        A convenience method for finding a specific hook.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a Hook object, instead of the
        generator.
        """
        client = client or Client()
        owner = owner or client.get_username()
        return next(
            Bitbucket(client=client).repositoryHookById(
                owner=owner,
                repository_name=repository_name,
                uuid=uuid))

    @staticmethod
    def find_hooks_for_repository(
            repository_name,
            owner=None,
            client=None):
        """
        A convenience method for finding hooks for a repository.
        The method is a generator Hooks objects.
        If no owner is provided, this method assumes client can provide one.
        """
        client = client or Client()
        owner = owner or client.get_username()
        return Bitbucket(client=client).repositoryHooks(
            owner=owner,
            repository_name=repository_name)


Client.bitbucket_types.add(Hook)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Provides a class for manipulating Team resources on Bitbucket.
"""

from pybitbucket.bitbucket import Bitbucket, BitbucketBase, Client, Enum


class TeamRole(Enum):
    ADMIN = 'admin'
    CONTRIBUTOR = 'contributor'
    MEMBER = 'member'


class Team(BitbucketBase):
    id_attribute = 'username'
    resource_type = 'teams'

    @staticmethod
    def is_type(data):
        return (Team.has_v2_self_url(data))

    @staticmethod
    def find_teams_for_role(role=TeamRole.ADMIN, client=Client()):
        """
        A convenience method for finding teams by the user's role.
        The method is a generator Team objects.
        """
        TeamRole(role)
        return Bitbucket(client=client).teamsForRole(role=role)

    @staticmethod
    def find_team_by_username(username, client=Client()):
        """
        A convenience method for finding a specific team.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a User object, instead of the
        generator.
        """
        return next(Bitbucket(client=client).teamByUsername(
            username=username))


Client.bitbucket_types.add(Team)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Provides classes for manipulating Comment resources.
"""
from uritemplate import expand

from pybitbucket.bitbucket import Bitbucket, BitbucketBase, Client


class Comment(BitbucketBase):
    id_attribute = 'id'
    resource_type = 'comments'

    @staticmethod
    def is_type(data):
        return (Comment.has_v2_self_url(data))

    # TODO: convert Comment to PayloadBuilder pattern.
    @staticmethod
    def make_payload(content):
        return {'content': {'raw': content}}

    @staticmethod
    def create_comment(
            content,
            snippet_id,
            username=None,
            client=Client()):
        if username is None:
            username = client.get_username()
        template = (
            '{+bitbucket_url}' +
            '/2.0/snippets/{username}/{snippet_id}' +
            '/comments')
        api_url = expand(
            template, {
                'bitbucket_url': client.get_bitbucket_url(),
                'username': username,
                'snippet_id': snippet_id,
            })
        payload = Comment.make_payload(content)
        response = client.session.post(api_url, data=payload)
        Client.expect_ok(response)
        return Comment(response.json(), client=client)

    @staticmethod
    def find_comment_for_snippet_by_id(
            snippet_id,
            comment_id,
            username=None,
            client=Client()):
        """
        A convenience method for finding a specific comment on a snippet.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a Comment object, instead of the
        generator.
        """
        if username is None:
            username = client.get_username()
        return next(Bitbucket(client=client).snippetCommentByCommentId(
            username=username,
            snippet_id=snippet_id,
            comment_id=comment_id))

    @staticmethod
    def find_comment_for_repository_commit_by_id(
            owner,
            repository_name,
            revision,
            comment_id,
            client=Client()):
        """
        A convenience method for finding a specific comment on a commit.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a Comment object, instead of the
        generator.
        """
        return next(
            Bitbucket(client=client).repositoryCommitCommentByCommentId(
                owner=owner,
                repository_name=repository_name,
                revision=revision,
                comment_id=comment_id))

    @staticmethod
    def find_comment_for_repository_pullrequest_by_id(
            owner,
            repository_name,
            pullrequest_id,
            comment_id,
            client=Client()):
        """
        A convenience method for finding a specific comment on a pull request.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a Comment object, instead of the
        generator.
        """
        return next(
            Bitbucket(client=client).repositoryPullRequestCommentsByCommentId(
                owner=owner,
                repository_name=repository_name,
                pullrequest_id=pullrequest_id,
                comment_id=comment_id))


Client.bitbucket_types.add(Comment)



# -*- coding: utf-8 -*-
import json
from os import path
from test_auth import FakeAuth

from pybitbucket.bitbucket import Client


class TestTypes(object):

    def setup_class(cls):
        cls.test_dir, current_file = path.split(path.abspath(__file__))
        cls.client = Client(FakeAuth())

    def object_from_file(self, filename):
        example_path = path.join(
            self.test_dir,
            filename)
        with open(example_path) as f:
            example = json.load(f)
        obj = self.client.convert_to_object(example)
        return obj

    def test_commit(self):
        import pybitbucket.commit  # noqa
        s = "%s" % self.object_from_file('Commit.json')
        assert s.startswith('Commit hash:')

    def test_repository(self):
        import pybitbucket.repository  # noqa
        s = "%s" % self.object_from_file('Repository.json')
        assert s.startswith('Repository full_name:')

    def test_snippet(self):
        import pybitbucket.snippet  # noqa
        s = "%s" % self.object_from_file('Snippet.json')
        assert s.startswith('Snippet id:')

    def test_team(self):
        import pybitbucket.team  # noqa
        s = "%s" % self.object_from_file('example_single_team.json')
        assert s.startswith('Team username:')

    def test_user(self):
        import pybitbucket.user  # noqa
        s = "%s" % self.object_from_file('User.json')
        assert s.startswith('User username:')



# -*- coding: utf-8 -*-
from test_bitbucketbase import BitbucketFixture
import json

import httpretty
from past.builtins import basestring
from uritemplate import expand
from pybitbucket.pullrequest import (
    PullRequest, PullRequestPayload, PullRequestState)
from pybitbucket.bitbucket import Bitbucket
from pybitbucket.comment import Comment
from pybitbucket.commit import Commit
from pybitbucket.repository import Repository
from pybitbucket.user import User
from voluptuous import MultipleInvalid


class PullRequestFixture(BitbucketFixture):
    # GIVEN: a class under test
    class_under_test = 'PullRequest'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return PullRequest(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes for a pull request
    pullrequest_id = 1
    title = 'Update entrypoint handling'
    description = ''
    state = PullRequestState.MERGED
    reason = ''
    reviewer = 'ian_buchanan'
    reviewers = [reviewer, 'tpettersen']
    close_source_branch = True
    source_repository_owner = 'pybitbucket'
    source_repository_name = 'snippet'
    source_repository_full_name = (
        source_repository_owner + '/' + source_repository_name)
    source_branch_name = 'feature-branch'
    source_commit = ''
    destination_branch_name = 'master'
    destination_commit = ''
    destination_repository_full_name = source_repository_full_name


class PullRequestPayloadFixture(PullRequestFixture):
    builder = PullRequestPayload()

    # GIVEN: a class under test
    class_under_test = 'PullRequestPayload'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return PullRequestPayload(json.loads(cls.resource_data()))


class TestGettingTheStringRepresentation(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.pullrequest_str = str(cls.example_object())

    def test_string_is_not_the_default_format(self):
        assert not self.pullrequest_str.startswith('<')
        assert not self.pullrequest_str.endswith('>')

    def test_string_has_the_class_name_and_id_attribute(self):
        assert self.pullrequest_str.startswith('PullRequest id:')


class TestCheckingTheExampleData(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.data = json.loads(cls.resource_data())

    def test_passes_the_type_check(self):
        assert PullRequest.is_type(self.data)


class TestAccessingPullRequestAttributes(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()

    def test_common_attributes_are_valid(self):
        assert self.pullrequest_id == self.response.id
        assert self.title == self.response.title
        assert self.response.description
        assert self.state == self.response.state
        assert self.reason == self.response.reason
        assert self.close_source_branch == self.response.close_source_branch

    def test_date_attributes(self):
        assert self.response.created_on
        assert self.response.updated_on

    def test_merge_commit_is_a_commit(self):
        assert isinstance(self.response.merge_commit, Commit)

    def test_destination_commit_is_a_commit(self):
        assert isinstance(self.response.destination_commit, Commit)

    def test_destination_repository_is_a_repository(self):
        assert isinstance(self.response.destination_repository, Repository)

    def test_source_commit_is_a_commit(self):
        assert isinstance(self.response.source_commit, Commit)

    def test_source_repository_is_a_repository(self):
        assert isinstance(self.response.source_repository, Repository)

    def test_closed_by_is_a_user(self):
        assert isinstance(self.response.closed_by, User)

    def test_author_is_a_user(self):
        assert isinstance(self.response.author, User)

    def test_reviewers_is_a_list_of_users(self):
        # TODO: reviewers array is missing sample data
        pass

    def test_participants_is_a_list_of_users(self):
        # TODO: participants array is missing sample data
        pass


class TestDeleting(PullRequestFixture):
    @httpretty.activate
    def test_response_is_not_an_exception(self):
        httpretty.register_uri(
            httpretty.DELETE,
            self.resource_url(),
            status=204)
        result = self.example_object().delete()
        assert result is None


class TestApproving(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.action_url = cls.get_link_url('approve')
        cls.response_data = cls.resource_data('PullRequest.approve')

    @httpretty.activate
    def test_approve_response_is_true(self):
        httpretty.register_uri(
            httpretty.POST,
            self.action_url,
            content_type='application/json',
            body=self.response_data,
            status=200)
        result = self.example_object().approve()
        assert result

    @httpretty.activate
    def test_unapprove_response_is_true(self):
        httpretty.register_uri(
            httpretty.DELETE,
            self.action_url,
            status=204)
        result = self.example_object().unapprove()
        assert result


class TestDeclining(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.action_url = cls.get_link_url('decline')
        cls.response_data = cls.resource_data('PullRequest.decline')

    @httpretty.activate
    def test_decline_response_is_true(self):
        httpretty.register_uri(
            httpretty.POST,
            self.action_url,
            content_type='application/json',
            body=self.response_data,
            status=200)
        result = self.example_object().decline()
        assert result


class TestMerging(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.action_url = cls.get_link_url('merge')
        cls.response_data = cls.resource_data('PullRequest.merge')

    @httpretty.activate
    def test_merge_response_is_true(self):
        httpretty.register_uri(
            httpretty.POST,
            self.action_url,
            content_type='application/json',
            body=self.response_data,
            status=200)
        result = self.example_object().merge()
        assert result


class TestCreatingNewPullRequest(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.url = expand(
            PullRequest.templates['create'], {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.test_client.get_username(),
                'repository_name': cls.source_repository_name,
            })

    @httpretty.activate
    def test_response_is_a_pullrequest(self):
        httpretty.register_uri(
            httpretty.POST,
            self.url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        payload = PullRequestPayload() \
            .add_title(self.title) \
            .add_source_branch_name(self.source_branch_name) \
            .add_source_repository_full_name(
                self.source_repository_full_name) \
            .add_destination_branch_name(self.destination_branch_name) \
            .add_destination_repository_full_name(
                self.destination_repository_full_name)
        response = PullRequest.create(payload, client=self.test_client)
        assert 'application/json' == \
            httpretty.last_request().headers.get('Content-Type')
        assert isinstance(response, PullRequest)


class TestFindingPullRequestById(PullRequestFixture):
    @httpretty.activate
    def test_response_is_a_pullrequest(self):
        httpretty.register_uri(
            httpretty.GET,
            self.resource_url(),
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = PullRequest.find_pullrequest_by_id_in_repository(
            pullrequest_id=self.pullrequest_id,
            repository_name=self.source_repository_name,
            owner=self.source_repository_owner,
            client=self.test_client)
        assert isinstance(response, PullRequest)


class TestFindingPullRequestsByState(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryPullRequestsInState', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.source_repository_owner,
                'repository_name': cls.source_repository_name,
                'state': cls.state
            })

    @httpretty.activate
    def test_response_from_only_name_is_a_repository_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = PullRequest.find_pullrequests_for_repository_by_state(
            repository_name=self.source_repository_name,
            client=self.test_client)
        assert isinstance(next(response), PullRequest)

    @httpretty.activate
    def test_response_from_all_parameters_is_a_repository_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = PullRequest.find_pullrequests_for_repository_by_state(
            repository_name=self.source_repository_name,
            owner=self.source_repository_owner,
            state=self.state,
            client=self.test_client)
        assert isinstance(next(response), PullRequest)


class TestAccessingLinks(PullRequestFixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()
        cls.commits_url = cls.get_link_url('commits')
        cls.commits_data = cls.resource_list_data('Commit')
        cls.comments_url = cls.get_link_url('comments')
        cls.comments_data = cls.resource_list_data('Comment')
        cls.diff_url = cls.get_link_url('diff')
        cls.diff_data = cls.data_from_file('Diff.txt')
        cls.activity_url = cls.get_link_url('activity')
        cls.activity_data = cls.resource_data('PullRequest.activity')

    @httpretty.activate
    def test_commits_returns_a_commit_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.commits_url,
            content_type='application/json',
            body=self.commits_data,
            status=200)
        response = self.response.commits()
        assert isinstance(next(response), Commit)

    @httpretty.activate
    def test_comments_returns_a_comments_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.comments_url,
            content_type='application/json',
            body=self.comments_data,
            status=200)
        print(self.commits_url)
        response = self.response.comments()
        assert isinstance(next(response), Comment)

    @httpretty.activate
    def test_diff_is_a_string(self):
        httpretty.register_uri(
            httpretty.GET,
            self.diff_url,
            content_type='application/json',
            body=self.diff_data,
            status=200)
        response = self.response.diff()
        assert isinstance(response, basestring)

    @httpretty.activate
    def test_activity_is_a_dictionary_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.activity_url,
            content_type='application/json',
            body=self.activity_data,
            status=200)
        response = self.response.activity()
        assert isinstance(next(response), dict)


class TestCreatingDefaultPullRequestPayload(PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.data = cls.builder.build()

    def test_default_payload_is_empty_dictionary(self):
        assert {} == self.data

    def test_default_is_invalid(self):
        try:
            self.builder.validate()
            assert False
        except Exception as e:
            assert isinstance(e, MultipleInvalid)


class TestAddingDestinationRepositoryToPullRequestPayload(
        PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.owner, cls.name = \
            cls.destination_repository_full_name.split('/', 1)
        cls.with_both = cls.builder \
            .add_destination_repository_owner(cls.owner) \
            .add_destination_repository_name(cls.name)
        cls.repository = Repository(
            json.loads(cls.resource_data('Repository')),
            client=cls.test_client)
        cls.with_repository = cls.builder \
            .add_destination_repository(cls.repository)

    def test_owner_and_name_match_both_inputs(self):
        assert self.owner == self.with_both.destination_repository_owner
        assert self.name == self.with_both.destination_repository_name

    def test_owner_and_name_match_object_input(self):
        assert 'teamsinspace' == \
            self.with_repository.destination_repository_owner
        assert 'teamsinspace.bitbucket.org' == \
            self.with_repository.destination_repository_name


class TestAddingTitleToPullRequestPayload(PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_title = cls.builder.add_title(cls.title)

    def test_immutability_on_adding_a_title(self):
        assert self.with_title
        assert {} == self.builder.build()

    def test_title_structure(self):
        expected = {
            'title': self.title
        }
        assert expected == self.with_title.build()


class TestAddingDestinationBranchNameToPullRequestPayload(
        PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_destination_branch_name = \
            cls.builder.add_destination_branch_name(
                cls.destination_branch_name)

    def test_immutability_on_adding_a_destination(self):
        assert self.with_destination_branch_name
        assert {} == self.builder.build()

    def test_destination_branch_name_structure(self):
        expected = {
            'destination': {
                'branch': {
                    'name': self.destination_branch_name
                }
            }
        }
        assert expected == self.with_destination_branch_name.build()


class TestAddingSourceBranchNameToPullRequestPayload(
        PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_source_branch_name = \
            cls.builder.add_source_branch_name(
                cls.source_branch_name)

    def test_immutability_on_adding_a_source(self):
        assert self.with_source_branch_name
        assert {} == self.builder.build()

    def test_source_branch_name_structure(self):
        expected = {
            'source': {
                'branch': {
                    'name': self.source_branch_name
                }
            }
        }
        assert expected == self.with_source_branch_name.build()


class TestAddingSourceRepositoryFullnameToPullRequestPayload(
        PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_source_repository_full_name = \
            cls.builder.add_source_repository_full_name(
                cls.source_repository_full_name)

    def test_immutability_on_adding_source_repository_full_name(self):
        assert self.with_source_repository_full_name
        assert {} == self.builder.build()

    def test_source_repository_full_name_structure(self):
        expected = {
            'source': {
                'repository': {
                    'full_name': self.source_repository_full_name
                }
            }
        }
        assert expected == self.with_source_repository_full_name.build()


class TestAddingDescriptionToPullRequestPayload(PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_description = cls.builder.add_description(cls.description)

    def test_immutability_on_adding_a_description(self):
        assert self.with_description
        assert {} == self.builder.build()

    def test_description_structure(self):
        expected = {
            'description': self.description
        }
        assert expected == self.with_description.build()


class TestAddingCloseSourceBranchToPullRequestPayload(
        PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_close_source_branch = \
            cls.builder.add_close_source_branch(
                cls.close_source_branch)

    def test_immutability_on_adding_close_source_branch(self):
        assert self.with_close_source_branch
        assert {} == self.builder.build()

    def test_close_source_branch_structure(self):
        expected = {
            'close_source_branch': self.close_source_branch
        }
        assert expected == self.with_close_source_branch.build()


class TestAddingReviewersToPullRequestPayload(PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_reviewer = \
            cls.builder.add_reviewer_by_username(
                cls.reviewer)
        cls.with_reviewers = \
            cls.builder.add_reviewers_from_usernames(
                cls.reviewers)
        cls.user = User(
            json.loads(cls.resource_data('User')),
            client=cls.test_client)
        cls.with_user_reviewer = cls.builder.add_reviewer(cls.user)

    def test_immutability_on_adding_reviewers(self):
        assert self.with_reviewer
        assert self.with_reviewers
        assert {} == self.builder.build()

    def test_reviewers_structure_for_one(self):
        expected = {
            'reviewers': [
                {'username': self.reviewer}
            ]
        }
        assert expected == self.with_reviewer.build()

    def test_reviewers_structure_for_many(self):
        expected = {'reviewers': [{'username': u} for u in self.reviewers]}
        assert expected == self.with_reviewers.build()

    def test_reviewers_structure_for_user(self):
        expected = {
            'reviewers': [
                {'username': 'evzijst'}
            ]
        }
        assert expected == self.with_user_reviewer.build()


class TestCreatingMinimalPullRequestPayload(PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = PullRequestPayload() \
            .add_title(cls.title) \
            .add_source_branch_name(cls.source_repository_name) \
            .add_source_repository_full_name(
                cls.source_repository_full_name) \
            .add_destination_branch_name(cls.destination_branch_name)
        cls.expected = json.loads(cls.resource_data(
            'PullRequestPayload.minimal'))

    def test_minimum_viable_payload_structure_for_create(self):
        assert self.payload.validate().build() == self.expected


class TestCreatingFullPullRequestPayload(PullRequestPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = PullRequestPayload() \
            .add_title('REQUIRED title') \
            .add_source_branch_name('REQUIRED name') \
            .add_source_repository_full_name('owner/repo_slug') \
            .add_destination_branch_name('name') \
            .add_destination_commit_by_hash('name') \
            .add_close_source_branch(True) \
            .add_description('description') \
            .add_reviewer_by_username('accountname')
        cls.expected = json.loads(cls.resource_data(
            'PullRequestPayload.full'))

    def test_full_payload_structure(self):
        assert self.payload.validate().build() == self.expected



# -*- coding: utf-8 -*-
from six import binary_type
import httpretty
from uritemplate import expand
from util import JsonSampleDataFixture
from pybitbucket.auth import (
    Authenticator,
    Anonymous, BasicAuthenticator,
    OAuth1Authenticator,
    OAuth2Grant, OAuth2Authenticator)


class FakeAuth(Authenticator):
    server_base_uri = 'https://staging.bitbucket.org/api'
    username = 'pybitbucket'
    password = 'secret'
    email = 'pybitbucket@mailinator.com'

    @classmethod
    def get_username(cls):
        return cls.username

    def __init__(self):
        self.session = self.start_http_session()


class AuthFixture(JsonSampleDataFixture):
    username = 'evzijst'
    email = 'pybitbucket@mailinator.com'
    server_base_uri = 'https://staging.bitbucket.org/api'
    auth = FakeAuth()

    @httpretty.activate
    def get_username_for_authenticator(self, auth):
        httpretty.HTTPretty.allow_net_connect = False
        httpretty.register_uri(
            httpretty.GET,
            auth.who_am_i_url,
            content_type='application/json',
            body=self.resource_data('User'),
            status=200)
        return auth.get_username()

    @staticmethod
    def safe_str(o):
        return o.decode('utf-8') if isinstance(o, binary_type) else o

    @httpretty.activate
    def get_request_headers(self, auth):
        httpretty.register_uri(httpretty.GET, self.server_base_uri)
        response = auth.session.get(self.server_base_uri)
        # Not sure why headers are sometimes unicode
        # and sometimes binary, but this solves it for testing.
        return {
            k: self.safe_str(v)
            for (k, v)
            in response.request.headers.items()}


class TestAuthFixture(AuthFixture):
    auth = FakeAuth()


class AnonymousFixture(AuthFixture):
    auth = Anonymous()


class BasicAuthenticatorFixture(AuthFixture):
    password = 'secret'
    # This digest has been precalculated for evzijst:secret
    digest = 'Basic ZXZ6aWpzdDpzZWNyZXQ='

    @classmethod
    def setup_class(cls):
        cls.auth = BasicAuthenticator(
            cls.username,
            cls.password,
            cls.email,
            server_base_uri=cls.server_base_uri)


class OAuth1AuthenticatorFixture(AuthFixture):
    client_key = '1'
    client_secret = 'secret'
    access_token = 'abc'
    access_token_secret = 'abc-secret'

    @classmethod
    def setup_class(cls):
        cls.auth = OAuth1Authenticator(
            cls.client_key,
            cls.client_secret,
            client_email=cls.email,
            server_base_uri=cls.server_base_uri)


class MockGrant(OAuth2Grant):
    callback_url = 'https://localhost'
    code = 'k9tSxZcUXRfnTqLH6J'
    response_template = '{+callback_url}/{?state,code}'

    def __init__(self):
        super(self.__class__, self).__init__()

    def obtain_authorization(self, session, auth_uri):
        authorization_url = session.authorization_url(auth_uri)
        print('Authorization URL: {}'.format(authorization_url[0]))
        redirect_response = expand(
            self.response_template, {
                'callback_url': self.callback_url,
                'state': authorization_url[1],
                'code': self.code,
            })
        return redirect_response


class OAuth2AuthenticatorFixture(AuthFixture):
    client_id = '1'
    client_secret = 'secret'
    redirect_uris = 'https://localhost'
    client_name = 'Test'
    client_description = 'Test Description'
    grant = MockGrant()

    @httpretty.activate
    def get_auth(self):
        self.token_uri = expand(
            '{+server_base_uri}/site/oauth2/access_token',
            {'server_base_uri': self.server_base_uri})
        token_response = """
{
"access_token":"2YotnFZFEjr1zCsicMWpAA",
"token_type":"bearer",
"expires_in":3600,
"refresh_token":"tGzv3JOkF0XG5Qx2TlKWIA",
"example_parameter":"example_value",
"scope":"repository"
}
        """
        httpretty.HTTPretty.allow_net_connect = False
        httpretty.register_uri(
            httpretty.POST,
            self.token_uri,
            content_type='application/json',
            body=token_response,
            status=200)
        a = OAuth2Authenticator(
            self.client_id,
            self.client_secret,
            self.email,
            self.grant,
            redirect_uris=self.redirect_uris,
            server_base_uri=self.server_base_uri,
            client_name=self.client_name,
            client_description=self.client_description)
        return a


class TestCreatingUserAgentHeaderString(AuthFixture):
    @classmethod
    def setup_class(cls):
        cls.user_agent_parts = Authenticator.user_agent_header().split(' ')

    def test_user_agent_starts_with_module_name(self):
        # It is most important that PyBitbucket is identified as the client.
        # While it is helpful to pass through the information provided
        # by the requests default, we don't have to test the contents.
        assert self.user_agent_parts[0].startswith('pybitbucket')


class TestCreatingHeaders(AuthFixture):
    @classmethod
    def setup_class(cls):
        cls.headers = Authenticator.headers()
        cls.headers_with_email = Authenticator.headers(email=cls.email)

    def test_accept_has_application_json(self):
        assert self.headers.get('Accept') == 'application/json'

    def test_user_agent_exists(self):
        # Already testing the contents separately, so just have one.
        assert self.headers.get('User-Agent')

    def test_default_has_from(self):
        # Default header doesn't know an email address.
        assert not self.headers.get('From')

    def test_email_is_in_from(self):
        assert self.headers_with_email.get('From') == self.email


class TestUsingTestAuth(TestAuthFixture):
    def test_constructor_was_able_to_construct_a_base_uri(self):
        assert self.auth.server_base_uri

    def test_constructor_has_created_an_http_session(self):
        assert self.auth.session

    def test_username_exists(self):
        assert self.auth.get_username() is not None

    def test_testauth_sends_no_authentication(self):
        assert not self.auth.session.auth
        assert not self.auth.session.headers.get('Authorization')


class TestUsingAnonymous(AnonymousFixture):
    def test_constructor_was_able_to_construct_a_base_uri(self):
        assert self.auth.server_base_uri

    def test_constructor_has_created_an_http_session(self):
        assert self.auth.session

    def test_anonymous_sends_no_authentication(self):
        assert not self.auth.session.auth
        assert not self.auth.session.headers.get('Authorization')

    def test_username_is_blank(self):
        # TODO: Should auth for anonymous return blank username?
        # Since the username is used in some methods
        # perhaps Anonymous should raise an exception?
        assert '' == self.auth.get_username()


class TestUsingBasicAuthentication(BasicAuthenticatorFixture):
    def test_constructor_was_able_to_construct_a_base_uri(self):
        assert self.auth.server_base_uri

    def test_constructor_has_created_an_http_session(self):
        assert self.auth.session

    def test_basicauth_has_authentication(self):
        assert self.auth.session.auth

    def test_username_exists(self):
        assert self.auth.get_username() is not None

    def test_username(self):
        assert self.username == self.auth.get_username()

    def test_sent_authorization_header(self):
        h = self.get_request_headers(self.auth)
        assert self.digest == h.get('Authorization')

    def test_sent_from_header(self):
        h = self.get_request_headers(self.auth)
        assert self.email == h.get('From')

    def test_sent_useragent_header(self):
        h = self.get_request_headers(self.auth)
        assert h.get('User-Agent').startswith('pybitbucket')

    def test_sent_accept_header(self):
        h = self.get_request_headers(self.auth)
        accept_params = h.get('Accept').split(';')
        json = [p for p in accept_params if p == 'application/json']
        assert any(json)


class TestUsingOAuth1Authentication(OAuth1AuthenticatorFixture):
    def test_constructor_was_able_to_construct_a_base_uri(self):
        assert self.auth.server_base_uri

    def test_constructor_has_created_an_http_session(self):
        assert self.auth.session

    def test_has_authentication(self):
        assert self.auth.session.auth

    def test_username(self):
        result = self.get_username_for_authenticator(self.auth)
        assert self.username == result

    def test_sent_authorization_header(self):
        h = self.get_request_headers(self.auth)
        assert h.get('Authorization')

    def test_sent_from_header(self):
        h = self.get_request_headers(self.auth)
        assert self.email == h.get('From')

    def test_sent_useragent_header(self):
        h = self.get_request_headers(self.auth)
        assert h.get('User-Agent').startswith('pybitbucket')

    def test_sent_accept_header(self):
        h = self.get_request_headers(self.auth)
        accept_params = h.get('Accept').split(';')
        json = [p for p in accept_params if p == 'application/json']
        assert any(json)


class TestUsingOAuth2Authentication(OAuth2AuthenticatorFixture):
    def test_constructor_was_able_to_construct_a_base_uri(self):
        assert self.get_auth().server_base_uri

    def test_constructor_has_created_an_http_session(self):
        assert self.get_auth().session

    def test_username(self):
        a = self.get_auth()
        result = self.get_username_for_authenticator(a)
        assert self.username == result

    def test_sent_authorization_header(self):
        a = self.get_auth()
        h = self.get_request_headers(a)
        assert h.get('Authorization')

    def test_sent_from_header(self):
        a = self.get_auth()
        h = self.get_request_headers(a)
        assert self.email == h.get('From')

    def test_sent_useragent_header(self):
        a = self.get_auth()
        h = self.get_request_headers(a)
        assert h.get('User-Agent').startswith('pybitbucket')

    def test_sent_accept_header(self):
        a = self.get_auth()
        h = self.get_request_headers(a)
        accept_params = h.get('Accept').split(';')
        json = [p for p in accept_params if p == 'application/json']
        assert any(json)



# -*- coding: utf-8 -*-
from test_bitbucketbase import BitbucketFixture
import json

import httpretty
from uritemplate import expand
from pybitbucket.hook import (Hook, HookPayload, HookEvent)
from pybitbucket.bitbucket import Bitbucket
from pybitbucket.repository import Repository


class HookFixture(BitbucketFixture):
    # GIVEN: a class under test
    class_under_test = 'Hook'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return Hook(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes for a hook
    owner = 'pybitbucket'
    repository_name = 'testing'
    hook_uuid = '{ec96bf6b-abb2-4f30-90b9-0178342c9fc5}'
    callback_url = 'https://example.com/bitbucket/'
    description = 'WebHook Description'
    active = True
    skip_cert_verification = False
    events = [HookEvent.REPOSITORY_PUSH]


class HookPayloadFixture(HookFixture):
    builder = HookPayload()

    # GIVEN: a class under test
    class_under_test = 'HookPayload'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return HookPayload(json.loads(cls.resource_data()))


class TestGettingTheStringRepresentation(HookFixture):
    @classmethod
    def setup_class(cls):
        cls.hook_str = str(cls.example_object())

    def test_string_is_not_the_default_format(self):
        assert not self.hook_str.startswith('<')
        assert not self.hook_str.endswith('>')

    def test_string_has_the_class_name_and_id_attribute(self):
        assert self.hook_str.startswith('Hook uuid:')


class TestCheckingTheExampleData(HookFixture):
    @classmethod
    def setup_class(cls):
        cls.data = json.loads(cls.resource_data())

    def test_passes_the_type_check(self):
        assert Hook.is_type(self.data)


class TestAccessingHookAttributes(HookFixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()

    def test_common_attributes_are_valid(self):
        assert self.hook_uuid == self.response.uuid
        assert self.callback_url == self.response.url
        assert self.description == self.response.description
        assert self.active == self.response.active
        assert self.skip_cert_verification == \
            self.response.skip_cert_verification

    def test_date_attributes(self):
        assert self.response.created_at

    def test_subject_is_a_repository(self):
        assert isinstance(self.response.subject, Repository)

    def test_events_are_valid(self):
        assert self.events == self.response.events


class TestCreatingNewHook(HookFixture):
    @classmethod
    def setup_class(cls):
        cls.url = expand(
            Hook.templates['create'], {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.test_client.get_username(),
                'repository_name': cls.repository_name,
            })

    @httpretty.activate
    def test_response_is_a_hook(self):
        httpretty.register_uri(
            httpretty.POST,
            self.url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        payload = HookPayload() \
            .add_description(self.description) \
            .add_callback_url(self.callback_url) \
            .add_events(self.events) \
            .activate() \
            .enable_cert_verification()
        response = Hook.create(
            payload=payload,
            repository_name=self.repository_name,
            owner=self.owner,
            client=self.test_client)
        assert 'application/json' == \
            httpretty.last_request().headers.get('Content-Type')
        assert isinstance(response, Hook)


class TestUpdating(HookFixture):
    new_url = 'http://www.example.com'

    @httpretty.activate
    def test_response_is_a_hook(self):
        httpretty.register_uri(
            httpretty.PUT,
            self.resource_url(),
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        payload = HookPayload() \
            .add_description(self.description) \
            .add_callback_url(self.new_url)
        response = self.example_object().update(payload=payload)
        assert self.new_url == \
            httpretty.last_request().parsed_body['url']
        assert isinstance(response, Hook)


class TestDeleting(HookFixture):
    @httpretty.activate
    def test_response_is_not_an_exception(self):
        httpretty.register_uri(
            httpretty.DELETE,
            self.resource_url(),
            status=204)
        result = self.example_object().delete()
        assert result is None


class TestFindingHookByUuid(HookFixture):
    @httpretty.activate
    def test_response_is_a_hook(self):
        httpretty.register_uri(
            httpretty.GET,
            self.resource_url(),
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = Hook.find_hook_by_uuid_in_repository(
            uuid=self.hook_uuid,
            repository_name=self.repository_name,
            owner=self.owner,
            client=self.test_client)
        assert isinstance(response, Hook)


class TestFindingHooksForRepository(HookFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryHooks', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.owner,
                'repository_name': cls.repository_name
            })

    @httpretty.activate
    def test_response_is_a_hook_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = Hook.find_hooks_for_repository(
            repository_name=self.repository_name,
            client=self.test_client)
        assert isinstance(next(response), Hook)


class TestCreatingMinimalHookPayload(HookPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = HookPayload() \
            .add_description(cls.description) \
            .add_callback_url(cls.callback_url)
        cls.expected = json.loads(cls.resource_data(
            'HookPayload.minimal'))

    def test_minimum_viable_payload_structure_for_create(self):
        assert self.payload.validate().build() == self.expected


class TestCreatingFullPullHookPayload(HookPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = HookPayload() \
            .add_description(cls.description) \
            .add_callback_url(cls.callback_url) \
            .add_events(cls.events) \
            .activate() \
            .enable_cert_verification()
        cls.expected = json.loads(cls.resource_data(
            'HookPayload.full'))

    def test_full_payload_structure(self):
        assert self.payload.validate().build() == self.expected



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from test_bitbucketbase import BitbucketFixture
import json

import httpretty
from uritemplate import expand
from pybitbucket.ref import (Ref, Tag, Branch)
from pybitbucket.bitbucket import Bitbucket
from pybitbucket.commit import Commit
from pybitbucket.repository import Repository


class RefFixture(BitbucketFixture):
    # GIVEN: a class under test
    class_under_test = 'Ref'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return Ref(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes for a refs
    owner = 'atlassian'
    repository_name = 'python-bitbucket'


class TagFixture(RefFixture):
    # GIVEN: a class under test
    class_under_test = 'Tag'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return Tag(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes for a tag
    name = 'v0.1.1'


class BranchFixture(RefFixture):
    # GIVEN: a class under test
    class_under_test = 'Branch'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return Branch(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes for a branch
    name = 'bug/35-url-name-collision'


class TestGettingTheStringRepresentationOfATag(TagFixture):
    @classmethod
    def setup_class(cls):
        cls.tag_str = str(cls.example_object())

    def test_string_is_not_the_default_format(self):
        assert not self.tag_str.startswith('<')
        assert not self.tag_str.endswith('>')

    def test_string_has_the_class_name_and_id_attribute(self):
        assert self.tag_str.startswith('Tag name:')


class TestGettingTheStringRepresentationOfABranch(BranchFixture):
    @classmethod
    def setup_class(cls):
        cls.branch_str = str(cls.example_object())

    def test_string_is_not_the_default_format(self):
        assert not self.branch_str.startswith('<')
        assert not self.branch_str.endswith('>')

    def test_string_has_the_class_name_and_id_attribute(self):
        assert self.branch_str.startswith('Branch name:')


class TestCheckingTheExampleTagData(TagFixture):
    @classmethod
    def setup_class(cls):
        cls.data = json.loads(cls.resource_data())

    def test_passes_the_type_check(self):
        assert Tag.is_type(self.data)


class TestCheckingTheExampleBranchData(BranchFixture):
    @classmethod
    def setup_class(cls):
        cls.data = json.loads(cls.resource_data())

    def test_passes_the_type_check(self):
        assert Branch.is_type(self.data)


class TestAccessingTagAttributes(TagFixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()

    def test_name_is_expected(self):
        assert self.name == self.response.name

    def test_target_is_a_commit(self):
        assert isinstance(self.response.target, Commit)


class TestAccessingBranchAttributes(BranchFixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()

    def test_name_is_expected(self):
        assert self.name == self.response.name

    def test_target_is_a_commit(self):
        assert isinstance(self.response.target, Commit)

    def test_repository_is_a_repository(self):
        assert isinstance(self.response.repository, Repository)


class TestDeletingATag(TagFixture):
    @httpretty.activate
    def test_response_is_not_an_exception(self):
        httpretty.register_uri(
            httpretty.DELETE,
            self.resource_url(),
            status=204)
        result = self.example_object().delete()
        assert result is None


class TestDeletingABranch(BranchFixture):
    @httpretty.activate
    def test_response_is_not_an_exception(self):
        httpretty.register_uri(
            httpretty.DELETE,
            self.resource_url(),
            status=204)
        result = self.example_object().delete()
        assert result is None


class TestFindingTagByNameInRepository(TagFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryTagByName', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.owner,
                'repository_name': cls.repository_name,
                'ref_name': cls.name
            })

    @httpretty.activate
    def test_response_is_a_tag(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = Tag.find_tag_by_ref_name_in_repository(
            ref_name=self.name,
            repository_name=self.repository_name,
            owner=self.owner,
            client=self.test_client)
        assert isinstance(response, Tag)


class TestFindingTagsInRepository(TagFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryTags', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.owner,
                'repository_name': cls.repository_name
            })

    @httpretty.activate
    def test_response_is_a_tag_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = Tag.find_tags_in_repository(
            repository_name=self.repository_name,
            owner=self.owner,
            client=self.test_client)
        assert isinstance(next(response), Tag)


class TestFindingTagsForSelf(TagFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryTags', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.test_client.get_username(),
                'repository_name': cls.repository_name
            })

    @httpretty.activate
    def test_response_from_only_repository_is_a_tag_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = Tag.find_tags_in_repository(
            repository_name=self.repository_name,
            client=self.test_client)
        assert isinstance(next(response), Tag)


class TestFindingBranchByNameInRepository(BranchFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryBranchByName', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.owner,
                'repository_name': cls.repository_name,
                'ref_name': cls.name
            })

    @httpretty.activate
    def test_response_is_a_branch(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = Branch.find_branch_by_ref_name_in_repository(
            ref_name=self.name,
            repository_name=self.repository_name,
            owner=self.owner,
            client=self.test_client)
        assert isinstance(response, Branch)


class TestFindingBranchInRepository(BranchFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryBranches', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.owner,
                'repository_name': cls.repository_name
            })

    @httpretty.activate
    def test_response_is_a_branch_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = Branch.find_branches_in_repository(
            repository_name=self.repository_name,
            owner=self.owner,
            client=self.test_client)
        assert isinstance(next(response), Branch)


class TestFindingBranchesForSelf(BranchFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('repositoryBranches', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'owner': cls.test_client.get_username(),
                'repository_name': cls.repository_name
            })

    @httpretty.activate
    def test_response_from_only_repository_is_a_branch_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = Branch.find_branches_in_repository(
            repository_name=self.repository_name,
            client=self.test_client)
        assert isinstance(next(response), Branch)


class TestAccessingTagLinks(TagFixture):
    @classmethod
    def get_link_url(cls, name):
        return (
            cls.response
            .data
            .get('links', {})
            .get(name, {})
            .get('href'))

    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()
        cls.commits_url = cls.get_link_url('commits')
        cls.commits_data = cls.resource_list_data('Commit')

    @httpretty.activate
    def test_commits_returns_a_commit_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.commits_url,
            content_type='application/json',
            body=self.commits_data,
            status=200)
        response = self.response.commits()
        assert isinstance(next(response), Commit)


class TestAccessingBranchLinks(BranchFixture):
    @classmethod
    def get_link_url(cls, name):
        return (
            cls.response
            .data
            .get('links', {})
            .get(name, {})
            .get('href'))

    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()
        cls.commits_url = cls.get_link_url('commits')
        cls.commits_data = cls.resource_list_data('Commit')

    @httpretty.activate
    def test_commits_returns_a_commit_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.commits_url,
            content_type='application/json',
            body=self.commits_data,
            status=200)
        response = self.response.commits()
        assert isinstance(next(response), Commit)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Defines the PullRequest resource and registers the type with the Client.

Classes:
- PullRequestState: enumerates the possible states of a pull request
- PullRequestPayload: encapsulates payload for creating
    and modifying pull requests
- PullRequest: represents a pull request for code review
"""

from functools import partial
from uritemplate import expand
from voluptuous import Schema, Required, Optional

from pybitbucket.bitbucket import (
        Bitbucket, BitbucketBase, Client, PayloadBuilder, Enum)


class PullRequestState(Enum):
    OPEN = 'OPEN'
    MERGED = 'MERGED'
    DECLINED = 'DECLINED'


class PullRequestPayload(PayloadBuilder):
    """
    A builder object to help create payloads
    for creating and updating pull requests.
    """

    schema = Schema({
        Required('title'): str,
        Optional('description'): str,
        Optional('close_source_branch'): bool,
        Optional('reviewers'): [{Required('username'): str}],
        Required('destination'): {
            Required('branch'): {
                Required('name'): str
            },
            Optional('commit'): {
                Required('hash'): str
            }
        },
        Required('source'): {
            Required('branch'): {
                Required('name'): str
            },
            Required('repository'): {
                Required('full_name'): str
            },
            Optional('commit'): {
                Required('hash'): str
            }
        }
    })

    def __init__(
            self,
            payload=None,
            destination_repository_owner=None,
            destination_repository_name=None):
        super(self.__class__, self).__init__(payload=payload)
        self._destination_repository_owner = destination_repository_owner
        self._destination_repository_name = destination_repository_name

    @property
    def destination_repository_owner(self):
        return self._destination_repository_owner

    @property
    def destination_repository_name(self):
        return self._destination_repository_name

    def add_title(self, title):
        new = self._payload.copy()
        new['title'] = title
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_description(self, description):
        new = self._payload.copy()
        new['description'] = description
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_close_source_branch(self, close):
        new = self._payload.copy()
        new['close_source_branch'] = close
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_reviewer_by_username(self, username):
        new = self._payload.copy()
        reviewers = self._payload.get('reviewers', [])
        if {'username': username} not in reviewers:
            reviewers.append({'username': username})
        new['reviewers'] = reviewers
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_reviewer(self, user):
        return self.add_reviewer_by_username(user.username)

    def add_reviewers_from_usernames(self, usernames):
        new = self._payload.copy()
        reviewers = self._payload.get('reviewers', [])
        for username in usernames:
            if {'username': username} not in reviewers:
                reviewers.append({'username': username})
        new['reviewers'] = reviewers
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_destination_repository_owner(self, owner):
        return PullRequestPayload(
            payload=self._payload.copy(),
            destination_repository_owner=owner,
            destination_repository_name=self._destination_repository_name)

    def add_destination_repository_name(self, name):
        return PullRequestPayload(
            payload=self._payload.copy(),
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=name)

    def add_destination_repository_full_name(self, full_name):
        owner, name = full_name.split('/', 1)
        return PullRequestPayload(
            payload=self._payload.copy(),
            destination_repository_owner=owner,
            destination_repository_name=name)

    def add_destination_repository(self, repository):
        owner, name = repository.full_name.split('/', 1)
        return PullRequestPayload(
            payload=self._payload.copy(),
            destination_repository_owner=owner,
            destination_repository_name=name)

    def add_destination_branch_name(self, name):
        new = self._payload.copy()
        destination = self._payload.get('destination', {})
        destination['branch'] = destination.get('branch', {})
        destination['branch']['name'] = name
        new['destination'] = destination
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_destination_branch(self, branch):
        owner, name = branch.repository.full_name.split('/', 1)
        return PullRequestPayload(
                payload=self._payload.copy(),
                destination_repository_owner=owner,
                destination_repository_name=name) \
            .add_destination_branch_name(branch.name)

    def add_destination_commit_by_hash(self, hash):
        new = self._payload.copy()
        destination = self._payload.get('destination', {})
        destination['commit'] = destination.get('commit', {})
        destination['commit']['hash'] = hash
        new['destination'] = destination
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_destination_commit(self, commit):
        owner, name = commit.repository.full_name.split('/', 1)
        return PullRequestPayload(
                payload=self._payload.copy(),
                destination_repository_owner=owner,
                destination_repository_name=name) \
            .add_destination_commit_by_hash(commit.hash)

    def add_source_branch_name(self, name):
        new = self._payload.copy()
        source = self._payload.get('source', {})
        source['branch'] = source.get('branch', {})
        source['branch']['name'] = name
        new['source'] = source
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_source_repository_full_name(self, full_name):
        new = self._payload.copy()
        source = self._payload.get('source', {})
        source['repository'] = source.get('repository', {})
        source['repository']['full_name'] = full_name
        new['source'] = source
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_source_branch(self, branch):
        return self.add_source_branch_name(branch.name) \
            .add_source_repository_full_name(branch.repository.full_name)

    def add_source_commit_by_hash(self, hash):
        new = self._payload.copy()
        source = self._payload.get('source', {})
        source['commit'] = source.get('commit', {})
        source['commit']['hash'] = hash
        new['source'] = source
        return PullRequestPayload(
            payload=new,
            destination_repository_owner=self._destination_repository_owner,
            destination_repository_name=self._destination_repository_name)

    def add_source_commit(self, commit):
        return self.add_source_commit_by_hash(commit.hash) \
            .add_source_repository_full_name(commit.repository.full_name)


class PullRequest(BitbucketBase):
    id_attribute = 'id'
    resource_type = 'pullrequests'
    templates = {
        'create': (
            '{+bitbucket_url}' +
            '/2.0/repositories' +
            '{/owner,repository_name}' +
            '/pullrequests')
    }

    @staticmethod
    def is_type(data):
        return (PullRequest.has_v2_self_url(data))

    def attr_from_subchild(self, target_attribute, child, child_object):
        if self.data.get(child, {}).get(child_object, {}):
            setattr(
                self,
                target_attribute,
                self.client.convert_to_object(
                    self.data[child][child_object]))

    def __init__(self, data, client=Client()):
        super(PullRequest, self).__init__(data, client=client)
        self.attr_from_subchild(
            'source_commit', 'source', 'commit')
        self.attr_from_subchild(
            'source_repository', 'source', 'repository')
        self.attr_from_subchild(
            'destination_commit', 'destination', 'commit')
        self.attr_from_subchild(
            'destination_repository', 'destination', 'repository')
        # Special treatment for approve, decline, merge, and diff
        if data.get('links', {}).get('approve', {}).get('href', {}):
            url = data['links']['approve']['href']
            # Approve is a POST on the approve link
            setattr(self, 'approve', partial(
                self.post_approval, template=url))
            # Unapprove is a DELETE on the approve link
            setattr(self, 'unapprove', partial(
                self.delete_approval, template=url))
        if data.get('links', {}).get('decline', {}).get('href', {}):
            url = data['links']['decline']['href']
            # Decline is a POST
            setattr(self, 'decline', partial(
                self.post, client=client, url=url, json=None))
        if data.get('links', {}).get('merge', {}).get('href', {}):
            url = data['links']['merge']['href']
            # Merge is a POST
            setattr(self, 'merge', partial(
                self.post, client=client, url=url, json=None))
        if data.get('links', {}).get('diff', {}).get('href', {}):
            url = data['links']['diff']['href']
            # Diff returns plain text
            setattr(self, 'diff', partial(
                self.content, url=url))

    def content(self, url):
        response = self.client.session.get(url)
        Client.expect_ok(response)
        return response.content

    @classmethod
    def create(
            cls,
            payload,
            repository_name=None,
            owner=None,
            client=None):
        """Create a new Pull Request.

        :param payload: the options for creating the new Pull Request
        :type payload: PullRequestPayload
        :param repository_name: name of the destination repository,
            also known as repo_slug. Optional, if provided in the payload.
        :type repository_name: str
        :param owner: the owner of the destination repository.
            Optional, if provided in the payload.
        :type owner: str
        :param client: the configured connection to Bitbucket.
            If not provided, assumes an Anonymous connection.
        :type client: bitbucket.Client
        :returns: the new repository object.
        :rtype: PullRequest
        :raises: ValueError
        """
        client = client or Client()
        owner = (
            owner or
            payload.destination_repository_owner)
        repository_name = (
            repository_name or
            payload.destination_repository_name)
        if not (owner and repository_name):
            raise ValueError('owner and repository_name are required')
        json = payload.validate().build()
        api_url = expand(
            cls.templates['create'], {
                'bitbucket_url': client.get_bitbucket_url(),
                'owner': owner,
                'repository_name': repository_name,
            })
        return cls.post(api_url, json=json, client=client)

    @staticmethod
    def find_pullrequest_by_id_in_repository(
            pullrequest_id,
            repository_name,
            owner=None,
            client=None):
        """
        A convenience method for finding a specific pull request.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a PullRequest object, instead of the
        generator.
        """
        client = client or Client()
        owner = owner or client.get_username()
        return next(
            Bitbucket(client=client).repositoryPullRequestByPullRequestId(
                owner=owner,
                repository_name=repository_name,
                pullrequest_id=pullrequest_id))

    @staticmethod
    def find_pullrequests_for_repository_by_state(
            repository_name,
            owner=None,
            state=None,
            client=None):
        """
        A convenience method for finding pull requests for a repository.
        The method is a generator PullRequest objects.
        If no owner is provided, this method assumes client can provide one.
        If no state is provided, the server will assume open pull requests.
        """
        client = client or Client()
        owner = owner or client.get_username()
        if (state is not None):
            PullRequestState(state)
        return Bitbucket(client=client).repositoryPullRequestsInState(
            owner=owner,
            repository_name=repository_name,
            state=state)


Client.bitbucket_types.add(PullRequest)



# -*- coding: utf-8 -*-
from test_bitbucketbase import BitbucketFixture
import json

import httpretty
from uritemplate import expand
from pybitbucket.bitbucket import Bitbucket, RepositoryType
from pybitbucket.snippet import (
    open_files, SnippetRole, Snippet, SnippetPayload)
from pybitbucket.user import User
from pybitbucket.comment import Comment
from pybitbucket.commit import Commit

# TODO: Fix TestSnippetPaging so it doesn't need following dependencies
from os import path
from test_auth import FakeAuth
from util import data_from_file
from pybitbucket.bitbucket import Client


class SnippetFixture(BitbucketFixture):
    # GIVEN: a class under test
    class_under_test = 'Snippet'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return Snippet(
            json.loads(cls.resource_data()),
            client=cls.test_client)

    # GIVEN: Example data attributes for a build status
    id = '9qa8b'
    is_private = True
    scm = RepositoryType.GIT
    title = 'My first snippet'
    filename = 'one.txt'
    owner = 'ianbuchanan'
    role = SnippetRole.OWNER


class SnippetPayloadFixture(SnippetFixture):
    builder = SnippetPayload()

    # GIVEN: a class under test
    class_under_test = 'SnippetPayload'

    # GIVEN: An example object created from example data
    @classmethod
    def example_object(cls):
        return SnippetPayload(json.loads(cls.resource_data()))


class TestGettingTheStringRepresentation(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.snippet_str = str(cls.example_object())

    def test_string_is_not_the_default_format(self):
        assert not self.snippet_str.startswith('<')
        assert not self.snippet_str.endswith('>')

    def test_string_has_the_class_name_and_id_attribute(self):
        assert self.snippet_str.startswith('Snippet id:')


class TestCheckingTheExampleData(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.data = json.loads(cls.resource_data())

    def test_passes_the_type_check(self):
        assert Snippet.is_type(self.data)


class TestAccessingSnippetAttributes(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()

    def test_common_attributes_are_valid(self):
        assert self.title == self.response.title
        assert self.scm == self.response.scm
        assert self.is_private == self.response.is_private

    def test_files_returns_filenames(self):
        # I would expect the JSON to be ordered
        # but somewhere that order was getting lost.
        filename = list(sorted(self.response.files))[0]
        assert self.filename == filename


class TestDeleting(SnippetFixture):
    @httpretty.activate
    def test_response_is_not_an_exception(self):
        httpretty.register_uri(
            httpretty.DELETE,
            self.resource_url(),
            status=204)
        result = self.example_object().delete()
        assert result is None


class TestCreatingNewSnippet(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.url = expand(
            Snippet.templates['create'], {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
            })
        example_upload = path.join(
            cls.test_dir(),
            'example_upload_1.txt')
        cls.files = open_files([example_upload])

    @httpretty.activate
    def test_response_is_a_snippet(self):
        httpretty.register_uri(
            httpretty.POST,
            self.url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        payload = SnippetPayload() \
            .add_title(self.title) \
            .add_scm(self.scm) \
            .add_is_private(self.is_private)
        response = Snippet.create(
            self.files,
            payload,
            client=self.test_client)
        content_type = httpretty.last_request().headers.get('Content-Type')
        assert content_type.startswith('multipart/form-data')
        assert isinstance(response, Snippet)


class TestCreatingNewSnippetWithTwoFiles(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.url = expand(
            Snippet.templates['create'], {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
            })
        example_upload_1 = path.join(
            cls.test_dir(),
            'example_upload_1.txt')
        example_upload_2 = path.join(
            cls.test_dir(),
            'example_upload_2.rst')
        cls.files = open_files([example_upload_1, example_upload_2])

    @httpretty.activate
    def test_response_is_a_snippet_with_two_files(self):
        httpretty.register_uri(
            httpretty.POST,
            self.url,
            content_type='application/json',
            body=self.resource_data('Snippet.two_files'),
            status=200)
        payload = SnippetPayload() \
            .add_title(self.title) \
            .add_scm(self.scm) \
            .add_is_private(self.is_private)
        response = Snippet.create(
            self.files,
            payload,
            client=self.test_client)
        content_type = httpretty.last_request().headers.get('Content-Type')
        assert content_type.startswith('multipart/form-data')
        assert isinstance(response, Snippet)
        # I would expect the JSON to be ordered
        # but somewhere that order was getting lost.
        expected = ['example_upload_1.txt', 'example_upload_2.rst']
        assert sorted(expected) == sorted(response.filenames)


class TestModifyingSnippet(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.url = cls.example_object().links['self']['href']

    @httpretty.activate
    def test_response_is_a_snippet(self):
        httpretty.register_uri(
            httpretty.PUT,
            self.url,
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        payload = SnippetPayload() \
            .add_title(self.title) \
            .add_scm(self.scm) \
            .add_is_private(self.is_private)
        response = self.example_object().modify(payload=payload)
        content_type = httpretty.last_request().headers.get('Content-Type')
        assert content_type.startswith('application/json')
        assert isinstance(response, Snippet)


class TestCreatingDefaultSnippetPayload(SnippetPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.data = cls.builder.build()

    def test_default_payload_is_empty_dictionary(self):
        assert {} == self.data

    def test_default_is_valid(self):
        assert self.builder.validate()


class TestAddingOwnerToSnippetPayload(SnippetPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_owner = cls.builder.add_owner(cls.owner)

    def test_immutability_after_add(self):
        assert self.with_owner
        assert {} == self.builder.build()

    def test_structure(self):
        assert self.owner == self.with_owner.owner


class TestAddingTitleToSnippetPayload(SnippetPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_title = cls.builder.add_title(cls.title)

    def test_immutability_after_add(self):
        assert self.with_title
        assert {} == self.builder.build()

    def test_structure(self):
        expected = {
            'title': self.title
        }
        assert expected == self.with_title.build()


class TestAddingScmToSnippetPayload(SnippetPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_scm = cls.builder.add_scm(cls.scm)

    def test_immutability_after_add(self):
        assert self.with_scm
        assert {} == self.builder.build()

    def test_structure(self):
        expected = {
            'scm': self.scm
        }
        assert expected == self.with_scm.build()


class TestAddingIsPrivateToSnippetPayload(SnippetPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.with_is_private = cls.builder.add_is_private(cls.is_private)

    def test_immutability_after_add(self):
        assert self.with_is_private
        assert {} == self.builder.build()

    def test_structure(self):
        expected = {
            'is_private': self.is_private
        }
        assert expected == self.with_is_private.build()


class TestCreatingFullSnippetPayload(SnippetPayloadFixture):
    @classmethod
    def setup_class(cls):
        cls.payload = SnippetPayload() \
            .add_title(cls.title) \
            .add_scm(cls.scm) \
            .add_is_private(cls.is_private)
        cls.expected = json.loads(cls.resource_data(
            'SnippetPayload.full'))

    def test_minimum_viable_payload_structure_for_create(self):
        assert self.payload.validate().build() == self.expected


class TestFindingSnippetById(SnippetFixture):
    @httpretty.activate
    def test_response_is_a_snippet(self):
        httpretty.register_uri(
            httpretty.GET,
            self.resource_url(),
            content_type='application/json',
            body=self.resource_data(),
            status=200)
        response = Snippet.find_snippet_by_id_and_owner(
            id=self.id,
            owner=self.owner,
            client=self.test_client)
        assert isinstance(response, Snippet)


class TestFindingSnippetsForRole(SnippetFixture):
    @classmethod
    def setup_class(cls):
        template = (
            Bitbucket(client=cls.test_client)
            .data
            .get('_links', {})
            .get('snippetsForRole', {})
            .get('href'))
        cls.url = expand(
            template, {
                'bitbucket_url': cls.test_client.get_bitbucket_url(),
                'role': cls.role
            })

    @httpretty.activate
    def test_response_is_a_snippet_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.url,
            content_type='application/json',
            body=self.resource_list_data(),
            status=200)
        response = Snippet.find_snippets_for_role(
            client=self.test_client)
        assert isinstance(next(response), Snippet)


class TestAccessingLinks(SnippetFixture):
    @classmethod
    def get_link_url(cls, name):
        return (
            cls.response
            .data
            .get('links', {})
            .get(name, {})
            .get('href'))

    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()
        cls.watchers_url = cls.get_link_url('watchers')
        cls.watchers_data = cls.resource_list_data('User')
        cls.commits_url = cls.get_link_url('commits')
        cls.commits_data = cls.resource_list_data('Commit')
        cls.comments_url = cls.get_link_url('comments')
        cls.comments_data = cls.resource_list_data('Comment')
        # TODO: Missing tests for patch and diff relationships on Snippet

    @httpretty.activate
    def test_watchers_returns_a_user_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.watchers_url,
            content_type='application/json',
            body=self.watchers_data,
            status=200)
        response = self.response.watchers()
        assert isinstance(next(response), User)

    @httpretty.activate
    def test_commits_returns_a_commit_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.commits_url,
            content_type='application/json',
            body=self.commits_data,
            status=200)
        response = self.response.commits()
        assert isinstance(next(response), Commit)

    @httpretty.activate
    def test_comments_returns_a_comment_generator(self):
        httpretty.register_uri(
            httpretty.GET,
            self.comments_url,
            content_type='application/json',
            body=self.comments_data,
            status=200)
        response = self.response.comments()
        assert isinstance(next(response), Comment)


class TestAccessingSnippetContent(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.response = cls.example_object()

    @httpretty.activate
    def test_content_is_a_string(self):
        url = self.response \
            .data['files'][self.filename]['links']['self']['href']
        httpretty.register_uri(
            httpretty.GET,
            url,
            body='example',
            status=200)
        content = self.response.content(self.filename)
        assert 'example' == content.decode('utf-8')

    def test_empty_content_for_missing_file(self):
        content = self.response.content(
            'this_file_is_not_in_the_snippet.test')
        assert not content


class TestOpeningFilesFromFilelist(SnippetFixture):
    @classmethod
    def setup_class(cls):
        cls.example_1 = path.join(
            cls.test_dir(),
            'example_upload_1.txt')
        cls.example_2 = path.join(
            cls.test_dir(),
            'example_upload_2.rst')
        cls.files = open_files([
            cls.example_1,
            cls.example_2])

    def test_list_has_two_files(self):
        assert 2 == len(self.files)

    def test_files_match_filenames(self):
        head, tail = self.files[0]
        assert 'file' == head
        filename, file = tail
        assert self.example_1 == filename


class TestSnippetPaging(object):
    @classmethod
    def setup_class(cls):
        cls.test_dir, current_file = path.split(path.abspath(__file__))
        cls.client = Client(FakeAuth())

    def load_example_snippet(self):
        example_path = path.join(
            self.test_dir,
            'Snippet.json')
        with open(example_path) as f:
            example = json.load(f)
        return Snippet(example, client=self.client)

    @httpretty.activate
    def test_snippet_list(self):
        url1 = (
            'https://api.bitbucket.org/2.0/snippets' +
            '?role=owner')
        example1 = data_from_file(
            self.test_dir,
            'example_snippets_page_1.json')
        httpretty.register_uri(
            httpretty.GET,
            url1,
            content_type='application/json',
            body=example1,
            status=200)
        snips = Snippet.find_snippets_for_role(
            SnippetRole.OWNER,
            client=self.client)
        snippet_list = []
        snippet_list.append(next(snips))
        snippet_list.append(next(snips))
        snippet_list.append(next(snips))

        url2 = (
            'https://' +
            'staging.bitbucket.org/api' +
            '/2.0/snippets' +
            '?role=owner&page=2')
        example2 = data_from_file(
            self.test_dir,
            'example_snippets_page_2.json')
        httpretty.register_uri(
            httpretty.GET,
            url2,
            content_type='application/json',
            body=example2,
            status=200)
        snippet_list.append(next(snips))
        snippet_list.append(next(snips))
        assert 5 == len(snippet_list)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Core classes for communicating with the Bitbucket API.

Classes:
- Enumeration: abstraction for a set of enumerated values
- Client: abstraction over HTTP requests to Bitbucket API
- BitbucketSpecialAction: an enum of special actions to be handled by children
- RepositoryType: an enum of repository types (Git, Hg)
- BitbucketBase: parent class for Bitbucket resources
- PayloadBuilder: parent class for payloads
- Bitbucket: root resource for the whole Bitbucket instance
- BadRequestError: exception wrapping bad HTTP requests
- ServerError: exception wrapping server errors
"""

from enum import Enum as EnumBase
from json import loads, dumps, JSONEncoder as JSONEncoderBase
from functools import partial
from requests import codes, models as requests_models
from requests.exceptions import HTTPError
from uritemplate import expand
from voluptuous import Schema

from pybitbucket.auth import Anonymous
from pybitbucket.entrypoints import entrypoints_json


# subclass Enum to make it behave the same way as the former custom Enum class
class Enum(EnumBase):
    def __eq__(self, o):
        return self is o or self.__class__(o).value == self.value


class JSONEncoder(JSONEncoderBase):
    def default(self, obj):
        if isinstance(obj, Enum):
            return obj.value
        return super(JSONEncoder, self).default(obj)


# monkey patch request's json handler
requests_models.complexjson.dumps = partial(dumps, cls=JSONEncoder)


class Client(object):
    bitbucket_types = set()

    @staticmethod
    def expect_ok(response, code=codes.ok):
        if code == response.status_code:
            return
        elif 400 == response.status_code:
            raise BadRequestError(response)
        elif 500 <= response.status_code:
            raise ServerError(response)
        else:
            response.raise_for_status()

    def convert_to_object(self, data):
        if isinstance(data, Enum):
            return data.value()
        for t in Client.bitbucket_types:
            if t.is_type(data):
                return t(data, client=self)
        return data

    def remote_relationship(self, template, **keywords):
        url = expand(template, keywords)
        while url:
            response = self.session.get(url)
            self.expect_ok(response)
            json_data = response.json()
            if isinstance(json_data, list):
                for item in json_data:
                    yield self.convert_to_object(item)
                url = None
            elif json_data.get('values'):
                for item in json_data['values']:
                    yield self.convert_to_object(item)
                url = json_data.get('next')
            else:
                yield self.convert_to_object(json_data)
                url = None

    def get_bitbucket_url(self):
        return self.config.server_base_uri

    def get_username(self):
        return self.config.get_username()

    def __init__(self, config=None):
        self.config = config or Anonymous()
        self.session = self.config.session


class BitbucketSpecialAction(Enum):
    APPROVE = 'approve'
    DECLINE = 'decline'
    MERGE = 'merge'
    DIFF = 'diff'


class RepositoryType(Enum):
    GIT = 'git'
    HG = 'hg'


class BitbucketBase(object):
    id_attribute = 'id'

    @staticmethod
    def expect_bool(name, value):
        if not isinstance(value, bool):
            raise TypeError(
                "{0} is {1} instead of bool".format(name, type(value)))

    @staticmethod
    def expect_list(name, value):
        if not isinstance(value, (list, tuple)):
            raise TypeError(
                "{0} is {1} instead of list".format(name, type(value)))

    @staticmethod
    def links_from(data):
        links = {}
        # Bitbucket doesn't currently use underscore.
        # HAL JSON does use underscore.
        for link_name in ('links', '_links'):
            if data.get(link_name):
                links.update(data.get(link_name))
        for name, body in links.items():
            # Ignore quirky Bitbucket clone link
            if isinstance(body, dict):
                for href, url in body.items():
                    if href == 'href':
                        yield (name, url)

    @staticmethod
    def _has_v2_self_url(data, resource_type, id_attribute):
        if (
                (data.get('links') is None) or
                (data['links'].get('self') is None) or
                (data['links']['self'].get('href') is None) or
                (data.get(id_attribute) is None)):
            return False
        # Since the structure is right, assume it is v2.
        is_v2 = True
        url_path = data['links']['self']['href'].split('/')
        # Start looking from the end of the path.
        position = -1
        # Since repos have a slash in the full_name,
        # we have to match as many parts as we find (1 or 2).
        # And sometimes the id is an integer.
        for id_part in str(data[id_attribute]).split('/')[::-1]:
            is_v2 = is_v2 and (id_part == url_path[position])
            position -= 1
        # After matching the id_attribute,
        # the resource_type should be the preceding part of the path.
        is_v2 = (resource_type == url_path[position])
        return is_v2

    @classmethod
    def has_v2_self_url(cls, data):
        return cls._has_v2_self_url(data, cls.resource_type, cls.id_attribute)

    def add_remote_relationship_methods(self, data):
        for name, url in BitbucketBase.links_from(data):
            if (name not in BitbucketSpecialAction):
                setattr(self, name, partial(
                    self.client.remote_relationship,
                    template=url))

    def add_inline_resources(self, data):
        for name, body in data.items():
            # author is not treated the same on all resources
            if name == 'author':
                # For Commits, author has a raw part and
                # a full User resource.
                if (body.get('raw') and body.get('user')):
                    setattr(self, 'raw_author', body['raw'])
                    setattr(self, 'author', self.client.convert_to_object(
                        body['user']))
                # For PullRequests, author is just a User resource.
                else:
                    setattr(self, name, self.client.convert_to_object(body))
            # If an attribute has a dictionary for a body,
            # then descend to check for embedded resources.
            elif isinstance(body, dict):
                setattr(self, name, self.client.convert_to_object(body))
            # If an attribute has a list for a body,
            # then descend into the array to check for embedded resources.
            elif isinstance(body, list):
                if (body and isinstance(body[0], dict)):
                    setattr(self, name, [
                        self.client.convert_to_object(i)
                        for i in body])
                else:
                    setattr(self, name, body)

    @classmethod
    def extract_templates_from_json(cls):
        """
        A helper method for 1.0 API resources that extracts uri templates
        found in links into a template that would be found
        on a 2.0 API resource.
        """
        links = loads(cls.links_json)
        return {
            name: url
            for (name, url)
            in cls.links_from(links)}

    @classmethod
    def expand_link_urls(cls, **kwargs):
        """
        A helper method for 1.0 API resources that expands uri templates
        found in links into a fully navigable URL as would be found
        with a HAL-JSON resource.
        """
        links = loads(cls.links_json)
        return {'_links': {
            name: {'href': expand(template, kwargs)}
            for (name, template)
            in cls.links_from(links)}}

    @classmethod
    def get_link_template(cls, name):
        """
        A helper method for 1.0 API resources that gets the raw uri template
        for a specific link.
        """
        links = loads(cls.links_json)
        templates = [v for k, v in cls.links_from(links) if k == name]
        return templates[0]

    def __init__(self, data, client=Client()):
        self.data = data
        self.client = client
        self.__dict__.update(data)
        self.add_remote_relationship_methods(data)
        self.add_inline_resources(data)

    def delete(self):
        url = self.links['self']['href']
        response = self.client.session.delete(url)
        # Deletes the resource and returns 204 (No Content).
        Client.expect_ok(response, 204)
        return

    def put(self, json=None, **kwargs):
        url = self.links['self']['href']
        response = self.client.session.put(url, json=json, **kwargs)
        Client.expect_ok(response)
        return self.client.convert_to_object(response.json())

    @staticmethod
    def post(url, json=None, data=None, client=Client(), **kwargs):
        if data:
            response = client.session.post(url, data=data, **kwargs)
        else:
            response = client.session.post(url, json=json, **kwargs)
        Client.expect_ok(response)
        return client.convert_to_object(response.json())

    def post_approval(self, template):
        response = self.client.session.post(template)
        Client.expect_ok(response)
        json_data = response.json()
        return json_data.get('approved')

    def delete_approval(self, template):
        response = self.client.session.delete(template)
        # Deletes the approval and returns 204 (No Content).
        Client.expect_ok(response, 204)
        return True

    def attributes(self):
        return list(self.data.keys())

    def relationships(self):
        return (
            list(self.data.get('_links', {}).keys()) +
            list(self.data.get('links', {}).keys()))

    def __repr__(self):
        return u'{name}({data})'.format(
            name=type(self).__name__,
            data=repr(self.data))

    def __str__(self):
        return u'{name} {id}:{data}'.format(
            name=type(self).__name__,
            id=self.id_attribute,
            data=getattr(self, self.id_attribute))


class PayloadBuilder(object):
    schema = Schema({})

    def __init__(self, payload=None):
        self._payload = payload or {}

    def build(self):
        payload = {}
        for k, v in self._payload.items():
            if isinstance(v, Enum):
                payload[k] = v.value
            else:
                payload[k] = v
        return payload

    def validate(self):
        self.schema(self._payload)
        return self


class Bitbucket(BitbucketBase):
    def __init__(self, client=Client()):
        self.data = loads(entrypoints_json)
        self.client = client
        self.add_remote_relationship_methods(self.data)


class BitbucketError(HTTPError):
    """Raise when Bitbucket has an HTTP error."""
    interpretation = "The client encountered an error."

    def format_message(self):
        return ('Attempted to request {url}.'
                '{interpretation} {code} - {text}').format(
                        url=self.url,
                        interpretation=self.interpretation,
                        code=self.code,
                        text=self.text)

    def __init__(self, response):
        self.url = response.url
        self.code = response.status_code
        self.text = response.text
        try:
            # if the response is json,
            # then make it part of the exception structure
            json_data = response.json()
            json_error_message = json_data.get('error').get('message')
            self.error_message = json_error_message
            self.__dict__.update(json_data)
        except ValueError:
            pass
        super(BitbucketError, self).__init__(
            self.format_message())


class BadRequestError(BitbucketError):
    """Raise when Bitbucket complains about a bad request."""
    interpretation = "Bitbucket considered it a bad request."

    def __init__(self, response):
        super(BadRequestError, self).__init__(response)


class ServerError(BitbucketError):
    """Raise when Bitbucket complains about a server error."""
    interpretation = "The client encountered a server error."

    def __init__(self, response):
        super(ServerError, self).__init__(response)

from pybitbucket.repository import Repository
from pybitbucket.branchrestriction import BranchRestriction
from pybitbucket.build import BuildStatus
from pybitbucket.comment import Comment
from pybitbucket.commit import Commit
from pybitbucket.consumer import Consumer
from pybitbucket.hook import Hook
from pybitbucket.pullrequest import PullRequest
from pybitbucket.ref import Ref
from pybitbucket.snippet import Snippet
from pybitbucket.team import Team
from pybitbucket.user import User



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

"""
Defines the Ref, Tag, and Branch resources
and registers those type with the Client.

Classes:
- Ref: represents the abstraction of a Git reference to a commit
- Tag: represents the tag resource that references a specific commit
- Branch: represents the branch resource that references a set of commits
"""
from pybitbucket.bitbucket import Bitbucket, BitbucketBase, Client


class Ref(BitbucketBase):
    id_attribute = 'name'

    @staticmethod
    def is_type(*args):
        # Ref is an abstraction so does not match on data
        return False

    @staticmethod
    def find_refs_in_repository(
            owner,
            repository_name,
            client=Client()):
        """
        A convenience method for finding refs in a repository.
        The method is a generator Ref subtypes of Tag and Branch.
        """
        return Bitbucket(client=client).repositoryRefs(
            owner=owner,
            repository_name=repository_name)


class Tag(Ref):
    resource_type = 'tags'

    @staticmethod
    def is_type(data):
        return (Tag.has_v2_self_url(data))

    @staticmethod
    def find_tags_in_repository(
            repository_name,
            owner=None,
            client=Client()):
        """
        A convenience method for finding tags in a repository.
        The method is a generator Tag objects.
        """
        owner = owner or client.get_username()
        return Bitbucket(client=client).repositoryTags(
            owner=owner,
            repository_name=repository_name)

    @staticmethod
    def find_tag_by_ref_name_in_repository(
            ref_name,
            repository_name,
            owner=None,
            client=Client()):
        """
        A convenience method for finding a specific tag.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a Tag object, instead of the
        generator.
        """
        owner = owner or client.get_username()
        return next(Bitbucket(client=client).repositoryTagByName(
            owner=owner,
            repository_name=repository_name,
            ref_name=ref_name))


class Branch(Ref):
    resource_type = 'branches'

    @staticmethod
    def is_type(data):
        return (Branch.has_v2_self_url(data))

    @staticmethod
    def find_branches_in_repository(
            repository_name,
            owner=None,
            client=Client()):
        """
        A convenience method for finding branches in a repository.
        The method is a generator Branch objects.
        """
        owner = owner or client.get_username()
        return Bitbucket(client=client).repositoryBranches(
            owner=owner,
            repository_name=repository_name)

    @staticmethod
    def find_branch_by_ref_name_in_repository(
            ref_name,
            repository_name,
            owner=None,
            client=Client()):
        """
        A convenience method for finding a specific branch.
        In contrast to the pure hypermedia driven method on the Bitbucket
        class, this method returns a Branch object, instead of the
        generator.
        """
        owner = owner or client.get_username()
        return next(Bitbucket(client=client).repositoryBranchByName(
            owner=owner,
            repository_name=repository_name,
            ref_name=ref_name))


Client.bitbucket_types.add(Ref)
Client.bitbucket_types.add(Tag)
Client.bitbucket_types.add(Branch)
